"""
This is a script for the frontend of the TreeModel class used to obtain Mote Carlo estimates of feature importance.

DEV: TODO features
a) TreeModel.fit: Add n_jobs as input
b) TreeModel.eval: Add n_features to output
"""
from typing import Optional, Dict, List, Tuple, Type, Union, Callable
from sklearn.base import ClassifierMixin, BaseEstimator
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
import pandas as pd
import numpy as np

import aaanalysis.utils as ut

from .backend.check_models import (check_match_list_model_classes_kwargs,
                                   check_match_labels_X,
                                   check_match_X_is_selected)
from .backend.tree_model.tree_model_fit import fit_tree_based_models
from .backend.tree_model.tree_model_predict_proba import monte_carlo_predict_proba
from .backend.tree_model.tree_model_eval import eval_feature_selections


# I Helper Functions
def check_is_preselected(is_preselected=None):
    """Check if preselected is valid"""
    is_preselected = ut.check_array_like(name="is_preselected_feature", val=is_preselected, accept_none=True,
                                         expected_dim=1, dtype="bool")
    if is_preselected is not None and sum(is_preselected) == 0:
        raise ValueError(f"Number of preselected features should not be 0 in 'is_preselected': {is_preselected}")
    return is_preselected


def check_n_feat_min_n_feat_max(n_feat_min=None, n_feat_max=None):
    """Check if vmin and vmax are valid numbers and vmin is less than vmax."""
    ut.check_number_range(name="n_feat_min", val=n_feat_min, min_val=1, just_int=True, accept_none=False)
    ut.check_number_range(name="n_feat_max", val=n_feat_max, min_val=1, just_int=True, accept_none=False)
    if n_feat_min > n_feat_max:
        raise ValueError(f"'n_feat_min' ({n_feat_min}) >= 'n_feat_max' ({n_feat_max}) not fulfilled.")


def check_n_cv(n_cv=None, labels=None):
    """Check if n_cv is valid"""
    ut.check_number_range(name="n_cv", val=n_cv, min_val=2, just_int=True)
    # Check if smaller than smaller class
    unique, counts = np.unique(labels, return_counts=True)
    min_class_count = min(counts)
    # Ensure n_cv is not greater than the smallest class count
    if n_cv > min_class_count:
        raise ValueError(f"'n_cv' should not be greater than the smallest class count ({min_class_count}), got {n_cv}.")


def check_metrics(name="list_eval_sores", metrics=None):
    """Check if evaluation metrics are valid"""
    valid_metrics = ["accuracy", "balanced_accuracy", "precision", "recall", "f1", "roc_auc"]
    if metrics is None:
        raise ValueError(f"'{name}' should not be None. Chose from these: {valid_metrics}")
    metrics = ut.check_list_like(name=name, val=metrics, accept_str=True)
    wrong_metrics = [x for x in metrics if x not in valid_metrics]
    if len(wrong_metrics) != 0:
        raise ValueError(f"Wrong 'metrics' ({wrong_metrics}). Chose from these: {valid_metrics}")


def check_list_is_selected(list_is_selected=None, X=None, convert_1d_to_2d=False):
    """Check if list_is_feature is list containing bool 2D matrix matching to X"""
    if not isinstance(list_is_selected, list):
        raise ValueError("'list_is_feature' must be a list with 2D boolean array.")
    list_is_selected = list_is_selected.copy()
    n_features = X.shape[1]
    str_error = f"Each element in 'list_is_feature' must be a 2D boolean array each with 'n_features'={n_features} elements."
    for index, is_feature_round in enumerate(list_is_selected):
        if convert_1d_to_2d and isinstance(is_feature_round, np.ndarray):
            if is_feature_round.ndim != 1:
                raise ValueError(f"Element at index {index} should be 1D boolean array if 'convert_1d_to_2d=True'")
            else:
                if is_feature_round.dtype != bool:
                    raise ValueError(f"Element at index {index} is not a boolean array.")
                if is_feature_round.size != n_features:
                    raise ValueError(f"1D array at index {index} does not match 'n_features'={n_features}.")
            list_is_selected[index] = is_feature_round.reshape(1, -1)  # Convert and update the array in the list

        is_feature_round = list_is_selected[index]
        if not isinstance(is_feature_round, np.ndarray):
            raise ValueError(f"Element at index {index} in 'list_is_feature' is not a numpy array.\n {str_error}")
        if is_feature_round.dtype != bool:
            raise ValueError(f"Element at index {index} in 'list_is_feature' is not a boolean array.\n {str_error}")
        if is_feature_round.ndim != 2:
            raise ValueError(f"Element at index {index} in 'list_is_feature' is not a 2D array.\n {str_error}")
        if is_feature_round.shape[1] != n_features:
            raise ValueError(
                f"Element at index {index} in 'list_is_feature' does not match the number of features in 'X'. "
                f"Expected {n_features} features, got {is_feature_round.shape[1]}.\n {str_error}")
    return list_is_selected


def check_match_df_feat_importance_arrays(df_feat=None, feat_importance=None, feat_importance_std=None, drop=False):
    """Check if df_feat matches with importance values"""
    if feat_importance is None:
        raise ValueError(f"'feat_importance' is None. Please fit TreeModel before adding feature importance.")
    if feat_importance_std is None:
        raise ValueError(f"'feat_importance_std' is None. Please fit TreeModel before adding feature importance.")
    n_feat = len(df_feat)
    n_feat_imp = len(feat_importance)
    n_feat_imp_std = len(feat_importance_std)
    if n_feat != n_feat_imp:
        raise ValueError(f"Mismatch of number of features in 'df_feat' (n={n_feat} and in 'feat_importance' (n={n_feat_imp})")
    if n_feat != n_feat_imp_std:
        raise ValueError(f"Mismatch of number of features in 'df_feat' (n={n_feat} and in 'feat_importance_std' (n={n_feat_imp})")
    if not drop:
        # Check if columns already exist
        if ut.COL_FEAT_IMPORT in list(df_feat):
            raise ValueError(f"'{ut.COL_FEAT_IMPORT}' already in 'df_feat' columns. To override, set 'drop=True'.")
        if ut.COL_FEAT_IMPORT_STD in list(df_feat):
            raise ValueError(f"'{ut.COL_FEAT_IMPORT_STD}' already in 'df_feat' columns. To override, set 'drop=True'.")



# II Main Functions
class TreeModel:
    """
    Tree Model class: A wrapper for tree-based models to obtain Monte Carlo estimates of feature
    importance and predictions [Breimann24c]_.

    Monte Carlo estimates are derived by averaging feature importance or prediction probabilities across various
    tree-based models and training rounds, enhancing the robustness and reproducibility of these estimates.
    Additionally, the class supports feature selection through recursive feature elimination (RFE) and offers
    comprehensive evaluation of feature selections.

    Attributes
    ----------
    list_models_ : Nested list with objects, shape (n_rounds, n_models)
        List with fitted tree-based models for every round after calling the ``fit`` method.
    feat_importance : array-like, shape (n_features)
        An arrays containing importance of each feature averaged across all rounds
        and trained models from `list_model_classes`.
    feat_importance_std : array-like, shape (n_features)
        An arrays containing standard deviation for feature importance across all rounds
        and trained models from `list_model_classes`. Same order as ``feature_importance``.
    is_selected_ : array-like, shape (n_rounds, n_features)
        2D array indicating features being selected by recursive features selection (True) or not (False) for each round.
        Same order as ``feature_importance``.
    """
    def __init__(self,
                 list_model_classes : List[Type[Union[ClassifierMixin, BaseEstimator]]] = None,
                 list_model_kwargs: Optional[List[Dict]] = None,
                 is_preselected: Optional[ut.ArrayLike1D] = None,
                 verbose: bool = True,
                 random_state: Optional[int] = None,
                 ):
        """
        Parameters
        ----------
        list_model_classes : list of Type[ClassifierMixin or BaseEstimator], default=[RandomForestClassifier, ExtraTreesClassifier]
            A list of tree-based model classes to be used for feature importance analysis.
        list_model_kwargs : list of dict, optional
            A list of dictionaries containing keyword arguments for each model in `list_model_classes`.
        is_preselected : array-like, shape (n_features)
            Boolean array indicating features being preselected before applying recursive features selection.
            ``True`` indicates that a feature is preselected and ``False`` that it is not.
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        Notes
        -----
        * All attributes are set during fitting via the :meth:`TreeModel.fit` method and can be directly accessed.

        See Also
        --------
        * :class:`sklearn.ensemble.RandomForestClassifier` for random forest model.
        * :class:`sklearn.ensemble.ExtraTreesClassifier` for extra trees model.

        Examples
        --------
        .. include:: examples/tree_model.rst
        """
        # Global parameters
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        is_preselected = check_is_preselected(is_preselected=is_preselected)
        # Model parameters
        if list_model_classes is None:
            list_model_classes = [RandomForestClassifier, ExtraTreesClassifier]
        elif not isinstance(list_model_classes, list):
            list_model_classes = [list_model_classes]   # Single models are possible as well (but not recommender)
        list_model_classes = ut.check_list_like(name="list_model_classes", val=list_model_classes,
                                                accept_none=False, min_len=1)
        ut.check_list_like(name="list_model_kwargs", val=list_model_kwargs, accept_none=True)
        if list_model_kwargs is None:
            list_model_kwargs = [{} for _ in list_model_classes]
        check_match_list_model_classes_kwargs(list_model_classes=list_model_classes, list_model_kwargs=list_model_kwargs)
        _list_model_kwargs = []
        for model_class, model_kwargs in zip(list_model_classes, list_model_kwargs):
            ut.check_mode_class(model_class=model_class)
            model_kwargs = ut.check_model_kwargs(model_class=model_class,
                                                 model_kwargs=model_kwargs,
                                                 attribute_to_check="feature_importances_",
                                                 random_state=random_state)
            _list_model_kwargs.append(model_kwargs)
        # Internal attributes
        self._verbose = verbose
        self._random_state = random_state
        self._list_model_classes = list_model_classes
        self._list_model_kwargs = _list_model_kwargs
        self._is_preselected = is_preselected
        # Output parameters (set during model fitting)
        self.feat_importance: Optional[ut.ArrayLike1D] = None
        self.feat_importance_std: Optional[ut.ArrayLike1D] = None
        self.is_selected_: Optional[ut.ArrayLike2D] = None
        self.list_models_: Optional[List[List[Union[ClassifierMixin, BaseEstimator]]]] = None

    def fit(self,
            X: ut.ArrayLike2D,
            labels: ut.ArrayLike1D = None,
            n_rounds: int = 5,
            use_rfe: bool = False,
            n_cv: int = 5,
            n_feat_min: int = 25,
            n_feat_max: int = 75,
            metric: str = "accuracy",
            step : Optional[int] = None,
            ) -> "TreeModel":
        """
        Fit tree-based models and compute average feature importance [Breimann24c]_.

        The feature importance is calculated across all models and rounds. In each round, the set of features can
        optionally be prefiltered using Recursive Feature Elimination (RFE) with a default RandomForestClassifier,
        if ``use_rfe`` is set to True. This RFE process iteratively reduces the number of features to enhance model
        performance, guided by the metric specified in ``metric``. The reduction continues until reaching
        ``n_feat_min``, with an upper limit of ``n_feat_max`` features considered.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        labels : array-like, shape (n_samples)
            Class labels for samples in ``X`` (typically, 1=positive, 0=negative).
        n_rounds : int, default=5
            The number of rounds (>=1) to fit the model.
        use_rfe : bool, default=False
            If ``True``, recursive feature elimination (RFE) is used with random forest model for feature selection.
        n_cv : int, default=5
            The number of cross-validation folds for RFE, must be > 1 and ≤ the smallest class's sample count.
        n_feat_min : int, default=25
            The minimum number of features to select each round for RFE. Should be 0 < ``n_feat_min`` <= ``n_feat_max``.
        n_feat_max : int, default=75
            The maximum number of features to select each round for RFE. Should be >= ``n_feat_min``.
        metric : str, default="accuracy"
            The name of the scoring function to use for cross-validation for RFE.
            Valid metrics are: {'accuracy', 'balanced_accuracy', 'precision', 'recall', 'f1', 'roc_auc'}
        step : int, optional
            Number of features to remove per RFE iteration. If ``None``, removes all features with the lowest importance
            scores each iteration, offering a faster but less precise approach. Should be < ``n_features``.
            Ideally, ``step`` should be much smaller than ``n_features``, with 1-5% of `n_features` being a recommended range.

        Returns
        -------
        TreeModel
            The fitted TreeModel instance.

        See Also
        --------
        * [Breimann24c]_ describes recursive feature elimination algorithm and feature importance aggregation.
        * :class:`sklearn.ensemble.RandomForestClassifier` for the random forest model used with default settings
          for recursive feature elimination.
        * :class:`sklearn.feature_selection.RFECV` for similar cross-validation based recursive feature elimination
          algorithm. This one does not provide an upper limit for the number of features to select.
        * :func:`sklearn.model_selection.cross_validate` for details on cross-validation
        * Sckit-learn `cross-validation <https://scikit-learn.org/stable/modules/cross_validation.html>`_ documentation.
        * Sckit-learn `classification metrics and scorings <https://scikit-learn.org/stable/modules/model_evaluation.html>`_.

        Examples
        --------
        .. include:: examples/tm_fit.rst
        """
        # Check input
        X = ut.check_X(X=X)
        n_samp, n_feat = X.shape
        ut.check_X_unique_samples(X=X, min_n_unique_samples=2)
        labels = check_match_labels_X(labels=labels, X=X)
        ut.check_number_range(name="n_rounds", val=n_rounds, min_val=1, just_int=True)
        ut.check_bool(name="use_rfe", val=use_rfe)
        check_n_cv(n_cv=n_cv, labels=labels)
        check_n_feat_min_n_feat_max(n_feat_min=n_feat_min, n_feat_max=n_feat_max)
        ut.check_str(name="metric", val=metric)
        check_metrics(name="metric", metrics=metric)
        ut.check_number_range(name="step", val=step, min_val=1, max_val=n_feat-1, accept_none=True, just_int=True)
        # Fit tree-based models and save feature importance
        args = dict(X=X, labels=labels,
                    list_model_classes=self._list_model_classes,
                    list_model_kwargs=self._list_model_kwargs,
                    is_preselected=self._is_preselected,
                    n_rounds=n_rounds, use_rfe=use_rfe, n_cv=n_cv,
                    n_feat_min=n_feat_min, n_feat_max=n_feat_max, metric=metric, step=step,
                    verbose=self._verbose, random_state=self._random_state)
        feature_importance, is_selected_rounds, list_models_rounds = fit_tree_based_models(**args)
        # Save results in output parameters
        self.feat_importance = np.mean(feature_importance, axis=0).round(5) * 100
        self.feat_importance_std = np.std(feature_importance, axis=0).round(5) * 100
        self.is_selected_ = is_selected_rounds
        self.list_models_ = list_models_rounds
        return self

    def eval(self,
             X: ut.ArrayLike2D,
             labels: ut.ArrayLike1D = None,
             list_is_selected: List[ut.ArrayLike2D] = None,
             convert_1d_to_2d: bool = False,
             names_feature_selections: Optional[List[str]] = None,
             n_cv: int = 5,
             list_metrics: Union[str, List[str]] = None,
             ) -> pd.DataFrame:
        """
        Evaluate the prediction performance for different feature selections.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        labels : array-like, shape (n_samples)
            Class labels for samples in ``X`` (typically, 1=positive, 0=negative).
        list_is_selected : array-like, shape (n_feature_sets, n_round, n_features)
            List of 2D boolean arrays with shape (n_rounds, n_features) indicating different feature selections.
        convert_1d_to_2d : bool, default=False
            If ``True``, convert all boolean arrays in `list_is_selected` from 1D to 2D arrays with a single row.
        names_feature_selections : list of str, optional
            List of dataset names corresponding to ``list_is_feature``.
        n_cv : int, default=5
            The number of cross-validation folds for RFE, must be > 1 and ≤ the smallest class's sample count.
        list_metrics : str or list of str, default=['accuracy', 'precision', 'recall', 'f1']
            List of scoring metrics to use for evaluation. Only metrics for binary classification are allowed.
            Valid metrics are: {'accuracy', 'balanced_accuracy', 'precision', 'recall', 'f1', 'roc_auc'}

        Returns
        -------
        df_eval : pd.DataFrame
            Evaluation results for feature subsets obtained by recursive feature selection give with ``list_is_feature``.

        Notes
        -----
        * :func:`sklearn.metrics.balanced_accuracy_score` is recommended if datasets are unbalanced.

        Warnings
        --------
        * ``list_metrics``: 'precision' and 'f1' metrics may trigger 'UndefinedMetricWarning' in imbalanced or
          small datasets due to division by zero if no positive predictions can be maid.

        See Also
        --------
        * Sckit-learn `classification metrics and scorings <https://scikit-learn.org/stable/modules/model_evaluation.html>`_.

        Examples
        --------
        .. include:: examples/tm_eval.rst
        """
        # Check input
        X = ut.check_X(X=X, min_n_unique_features=2)
        n_samples, n_feat = X.shape
        ut.check_X_unique_samples(X=X, min_n_unique_samples=2)
        labels = check_match_labels_X(labels=labels, X=X)
        ut.check_list_like(name="list_is_selected", val=list_is_selected)
        ut.check_bool(name="convert_1d_to_2d", val=convert_1d_to_2d)
        list_is_selected = check_list_is_selected(list_is_selected=list_is_selected, X=X,
                                                  convert_1d_to_2d=convert_1d_to_2d)
        names_feature_selections = ut.check_list_like(name="name_feature_selections", val=names_feature_selections, accept_none=True,
                                                      accept_str=True, check_all_str_or_convertible=True)
        if list_metrics is None:
            list_metrics = ["accuracy", "precision", "recall", "f1"]
        check_metrics(name="list_metrics", metrics=list_metrics)
        check_n_cv(n_cv=n_cv, labels=labels)
        # Perform evaluation
        df_eval = eval_feature_selections(X, labels=labels,
                                          list_is_selected=list_is_selected,
                                          names_feature_selections=names_feature_selections,
                                          n_cv=n_cv,
                                          list_metrics=list_metrics,
                                          list_model_classes=self._list_model_classes,
                                          list_model_kwargs=self._list_model_kwargs,
                                          verbose=self._verbose)

        return df_eval

    def predict_proba(self,
                      X: ut.ArrayLike2D,
                      ) -> Tuple[ut.ArrayLike1D, ut.ArrayLike1D]:
        """
        Obtain Monte Carlo estimate of class prediction probabilities for the positive class in `X`.

        Predictions are performed using all tree-based models from the `list_model_classes` attribute and
        feature selections from the `is_selected` attribute.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. Rows typically correspond to samples and columns to features.

        Returns
        -------
        pred : array-like, shape (n_samples)
            Array with the average prediction score for the positive class for each sample.
        pred_std : array-like, shape (n_samples)
            Array with the standard deviation of prediction scores for the positive class for each sample.

        See Also
        --------
        * :class:`sklearn.ensemble.RandomForestClassifier` for information on tree-based models.
        * :meth:`TreeModel.fit` for fitting the TreeModel object.

        Examples
        --------
        .. include:: examples/tm_predict_proba.rst
        """
        # Check input
        X = ut.check_X(X=X)
        check_match_X_is_selected(X=X, is_selected=self.is_selected_)
        # Predictions
        pred, pred_std = monte_carlo_predict_proba(X=X,
                                                   list_models=self.list_models_,
                                                   is_selected=self.is_selected_)
        return pred, pred_std

    def add_feat_importance(self,
                            df_feat: pd.DataFrame = None,
                            drop: bool = False
                            ) -> pd.DataFrame:
        """
        Include feature importance and its standard deviation to feature DataFrame.

        Feature importance is included as ``feat_importance`` column and the standard deviation of
        the feature importance as ``feat_importance_std`` column.

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
        drop : bool, default=False
            If ``True``, allow dropping of already existing ``feat_importance`` and ``feat_importance_std`` columns
            from ``df_feat`` before inserting.

        Returns
        -------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info+2)
            Feature DataFrame including ``feat_importance`` and ``feat_importance_std`` columns.

        See Also
        --------
        * :meth:`CPP.run` for details on CPP statistical measures of feature DataFrame.

        Examples
        --------
        .. include:: examples/tm_add_feat_importance.rst
         """
        # Check input
        df_feat = ut.check_df_feat(df_feat=df_feat)
        ut.check_bool(name="drop", val=drop)
        check_match_df_feat_importance_arrays(df_feat=df_feat,
                                              feat_importance=self.feat_importance,
                                              feat_importance_std=self.feat_importance_std,
                                              drop=drop)
        # Add feature importance and its standard deviation to the DataFrame
        df_feat = df_feat.copy()
        if drop:
            columns = [x for x in list(df_feat) if x not in [ut.COL_FEAT_IMPORT, ut.COL_FEAT_IMPORT_STD]]
            df_feat = df_feat[columns]
        args = dict(allow_duplicates=False)
        df_feat.insert(loc=len(df_feat.columns), column=ut.COL_FEAT_IMPORT, value=self.feat_importance, **args)
        df_feat.insert(loc=len(df_feat.columns), column=ut.COL_FEAT_IMPORT_STD, value=self.feat_importance_std, **args)
        return df_feat



"""
This is a script for the backend PlotPosition utility class for the CPPPlot class.
"""
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl

import aaanalysis.utils as ut
from .utils_feature import get_positions_, get_df_pos_


# I Helper Functions
# Optimize fontsize
def _get_optimal_fontsize(ax=None, labels=None, max_x_dist=0.1):
    """Optimize font size of sequence characters"""
    min_fontsize, max_fontsize = 1, 60
    th_binary_search = 0.01
    fs_reduction = 0.05
    # Line width for the bounding box
    lw = ax.figure.get_size_inches()[0] / 5
    # Function to compute the bounding box for each label
    f = lambda l: l.get_window_extent(ax.figure.canvas.get_renderer()).transformed(ax.transData.inverted())
    def set_label_properties(_label, _fontsize):
        _label.set_fontsize(_fontsize)
        _label.set_bbox(dict(facecolor='none', edgecolor='none', zorder=0.1, alpha=1, clip_on=False, pad=0, linewidth=lw))

    # Function to check overlap between bounding boxes
    def check_overlap(_bbox_list):
        for i in range(len(_bbox_list) - 1):
            x_distance = _bbox_list[i + 1].x0 - _bbox_list[i].x1
            if x_distance < max_x_dist:
                return True
        return False

    # Binary search to narrow down the font size range
    while max_fontsize - min_fontsize > th_binary_search:
        fontsize = (min_fontsize + max_fontsize) / 2
        for label in labels:
            set_label_properties(label, fontsize)
        bbox_list = [f(label) for label in labels]
        if check_overlap(bbox_list):
            max_fontsize = fontsize
        else:
            min_fontsize = fontsize

    # High precision step-wise reduction
    optimal_fontsize = max_fontsize
    while optimal_fontsize > min_fontsize:
        for label in labels:
            set_label_properties(label, optimal_fontsize)
        bbox_list = [f(label) for label in labels]
        if not check_overlap(bbox_list):
            break  # If no overlap is detected, this is the optimal size
        optimal_fontsize -= fs_reduction
    return optimal_fontsize


# Positions
def _get_df_pos_sign(df_feat=None, count=True, col_cat="category", col_val=None, value_type="count",
                     start=None, stop=None):
    """Get DataFrame for positive and negative values based on feature importance."""
    kwargs = dict(col_cat=col_cat, col_val=col_val, value_type=value_type,
                  start=start, stop=stop)
    list_df = []
    df_p = df_feat[df_feat[col_val] > 0]
    df_n = df_feat[df_feat[col_val] <= 0]
    if len(df_p) > 0:
        df_positive = get_df_pos_(df_feat=df_p, **kwargs)
        list_df.append(df_positive)
    if len(df_n) > 0:
        df_negative = get_df_pos_(df_feat=df_n, **kwargs)
        if count:
            df_negative = -df_negative
        list_df.append(df_negative)
    df_pos = pd.concat(list_df)
    return df_pos


# TMD-JMD bar functions
def _adjust_xticks_labels(xticks=None, xtick_labels=None, add_xtick_pos=True,
                          exists_jmd_n=True, exists_jmd_c=True):
    """Remove JMD-N and/or JMD-C from x-ticks and x-tick labels if not exist"""
    n = 2 if add_xtick_pos else 1
    if not exists_jmd_n:
        # Remove JMD-N related ticks and labels if it does not exist
        xticks = xticks[n:]
        xtick_labels = xtick_labels[n:]
    if not exists_jmd_c:
        # Remove JMD-C related ticks and labels if it does not exist
        xticks = xticks[:-n]
        xtick_labels = xtick_labels[:-n]
    return xticks, xtick_labels


def _get_bar_height(ax=None, divider=50):
    """Calculate bar height for sequence part visualization."""
    ylim = ax.get_ylim()
    width, height = plt.gcf().get_size_inches()
    bar_height = abs(ylim[0] - ylim[1]) / divider / height*6
    return bar_height


def _get_y(ax=None, bar_height=None, height_factor=1.0, reversed_weight=0):
    """Determine y-coordinate for bar placement."""
    ylim = ax.get_ylim()
    reversed_y = reversed_weight if ylim[0] > ylim[1] else 1
    y = ylim[0] - (bar_height * height_factor) * reversed_y
    return y


def _add_part_bar(ax=None, start=1.0, len_part=40.0, color="blue"):
    """Add colored bar for TMD and JMD sequence parts."""
    bar_height = _get_bar_height(ax=ax)
    y = _get_y(ax=ax, bar_height=bar_height)
    bar = mpl.patches.Rectangle((start, y), width=len_part, height=bar_height, linewidth=0,
                                color=color, zorder=3, clip_on=False)
    ax.add_patch(bar)


def _add_part_text(ax=None, text=None, start=1.0, len_part=10.0, fontsize=None, fontweight="normal"):
    """Place text marking for TMD and JMD sequence parts."""
    bar_height = _get_bar_height(ax=ax)
    y = _get_y(ax=ax, bar_height=bar_height, height_factor=1.3, reversed_weight=-1)
    x = start + len_part / 2    # Middle of part
    ax.text(x, y, text,
            horizontalalignment='center',
            verticalalignment='top',
            fontsize=fontsize,
            fontweight=fontweight,
            color='black')


def _add_part_seq(ax=None, jmd_n_seq=None, tmd_seq=None, jmd_c_seq=None, x_shift=0.0, seq_size=None,
                  tmd_color="mediumspringgreen", jmd_color="blue", tmd_seq_color="black", jmd_seq_color="white"):
    """Add colored boxes for TMD and JMD sequences."""
    tmd_jmd = jmd_n_seq + tmd_seq + jmd_c_seq
    colors = [jmd_color] * len(jmd_n_seq) + [tmd_color] * len(tmd_seq) + [jmd_color] * len(jmd_c_seq)
    dict_seq_color = {tmd_color: tmd_seq_color, jmd_color: jmd_seq_color}
    xticks = range(0, len(tmd_jmd))
    # Set major and minor ticks to enable proper grid lines
    major_xticks = [x for x in xticks if x in ax.get_xticks()]
    minor_xticks = [x for x in xticks if x not in ax.get_xticks()]
    ax.set_xticks([x + x_shift for x in major_xticks], minor=False)
    ax.set_xticks([x + x_shift for x in minor_xticks], minor=True)
    kws_ticks = dict(rotation="horizontal", fontsize=seq_size, fontweight="bold",
                     fontname=ut.FONT_AA, zorder=2)
    ax.set_xticklabels(labels=[tmd_jmd[i] for i in minor_xticks], minor=True, **kws_ticks)
    ax.set_xticklabels(labels=[tmd_jmd[i] for i in major_xticks], minor=False, **kws_ticks)
    ax.tick_params(axis="x", length=0, which="both")
    # Get labels in order of sequence (separate between minor and major ticks)
    dict_pos_label = dict(zip(minor_xticks, ax.xaxis.get_ticklabels(which="minor")))
    dict_pos_label.update(dict(zip(major_xticks, ax.xaxis.get_ticklabels(which="major"))))
    labels = list(dict(sorted(dict_pos_label.items(), key=lambda item: item[0])).values())
    # Adjust font size to prevent overlap
    if seq_size is None:
        seq_size = _get_optimal_fontsize(ax=ax, labels=labels)
    # Set colored box around sequence to indicate TMD, JMD
    lw = plt.gcf().get_size_inches()[0]/5
    for l, c in zip(labels, colors):
        l.set_fontsize(seq_size)
        l.set_bbox(dict(facecolor=c, edgecolor=c, zorder=0.1, alpha=1, clip_on=False,
                        pad=0, linewidth=lw))
        l.set_color(dict_seq_color[c])
    return seq_size


def _add_part_seq_second_ticks(ax2=None, seq_size=11.0, xticks=None, xtick_labels=None,
                               fontsize_tmd_jmd=11, weight_tmd_jmd="normal",
                               xtick_size=11, x_shift=0.5, heatmap=False):
    """Add additional ticks for box of sequence parts"""
    name_tmd = ut.options["name_tmd"]
    name_jmd_n = ut.options["name_jmd_n"]
    name_jmd_c = ut.options["name_jmd_c"]
    # Move twinned axis ticks and label from top to bottom
    ax2.xaxis.set_ticks_position("bottom")
    ax2.xaxis.set_label_position("bottom")
    # Offset the twin axis below the host
    y_pos = 5 if heatmap else 7
    ax2.spines["bottom"].set_position(("outward", seq_size+y_pos))
    ax2.set_xticks([x + x_shift for x in xticks])
    ax2.set_xticklabels(xtick_labels, size=xtick_size, rotation=0, color="black", ha="center", va="top")
    ax2.tick_params(axis="x", color="black", length=0, width=0, bottom=True, pad=0)
    ax2.set_frame_on(False)
    labels = ax2.xaxis.get_ticklabels()
    for l in labels:
        text = l.get_text()
        if name_tmd in text:
            l.set_size(fontsize_tmd_jmd)
            l.set_weight(weight_tmd_jmd)
        elif name_jmd_n in text or name_jmd_c in text:
            l.set_size(fontsize_tmd_jmd)
            l.set_weight(weight_tmd_jmd)


def _get_new_axis(ax=None):
    """Get new axis object with same y-axis as input ax"""
    ax_new = ax.figure.add_subplot(ax.get_subplotspec(), frameon=False, yticks=[])
    return ax_new


# II Main Functions
class PlotPartPositions:
    """Class for plotting positional information for CPP analysis"""

    def __init__(self, tmd_len=20, jmd_n_len=10, jmd_c_len=10, start=1):
        """Initialize the plot positions with given lengths and start position."""
        self.jmd_n_len = jmd_n_len
        self.tmd_len = tmd_len
        self.jmd_c_len = jmd_c_len
        self.seq_len = jmd_n_len + tmd_len + jmd_c_len
        self.start = start
        self.stop = start + self.seq_len - 1

    # Helper methods
    def _get_starts(self, x_shift=0):
        """Calculate the starting positions for JMD and TMD."""
        jmd_n_start = 0 + x_shift
        tmd_start = self.jmd_n_len + x_shift
        jmd_c_start = self.jmd_n_len + self.tmd_len + x_shift
        return jmd_n_start, tmd_start, jmd_c_start

    def _get_middles(self, x_shift=0.0):
        """Calculate the middle positions for JMD and TMD."""
        jmd_n_middle = int(self.jmd_n_len/2) + x_shift
        tmd_middle = self.jmd_n_len + int(self.tmd_len/2) + x_shift
        jmd_c_middle = self.jmd_n_len + self.tmd_len + int(self.jmd_c_len/2) + x_shift
        return jmd_n_middle, tmd_middle, jmd_c_middle

    def _get_ends(self, x_shift=-1):
        """Calculate the ending positions for JMD and TMD."""
        jmd_n_end = self.jmd_n_len + x_shift
        tmd_end = self.jmd_n_len + self.tmd_len + x_shift
        jmd_c_end = self.jmd_n_len + self.tmd_len + self.jmd_c_len + x_shift
        return jmd_n_end, tmd_end, jmd_c_end

    # Main methods
    def get_df_pos(self, df_feat=None, df_cat=None,
                   col_cat="category", col_val="mean_dif",
                   value_type="count", normalize=False):
        """Get df_pos with values (e.g., counts or mean auc) for each feature (y) and positions (x)"""
        df_feat = df_feat.copy()
        # Adjust feature positions
        features = df_feat[ut.COL_FEATURE].to_list()
        feat_positions = get_positions_(features=features,
                                        start=self.start,
                                        tmd_len=self.tmd_len,
                                        jmd_n_len=self.jmd_n_len,
                                        jmd_c_len=self.jmd_c_len)
        df_feat[ut.COL_POSITION] = feat_positions
        # Get dataframe with positions
        kwargs = dict(col_cat=col_cat, col_val=col_val, value_type=value_type, start=self.start, stop=self.stop)
        if value_type == "count":
            if col_val is None:
                df_pos = get_df_pos_(df_feat=df_feat, **kwargs)
            else:
                df_pos = _get_df_pos_sign(df_feat=df_feat, count=True, **kwargs)
        else:
            df_pos = get_df_pos_(df_feat=df_feat, **kwargs)
        if normalize:
            df_pos = df_pos / abs(df_pos).sum().sum() * 100
        # Sort according to given categories
        list_cat = list(df_cat[col_cat].drop_duplicates())
        list_col = list(df_pos.T)
        sorted_col = [x for x in list_cat if x in list_col]
        df_pos = df_pos.T[sorted_col].T
        return df_pos.round(3)

    # Add TMD-JMD sequence
    @staticmethod
    def get_optimal_fontsize(ax, labels, max_x_dist=0.1):
        """Get sequence fontsize optimized to not overlap"""
        opt_fs = _get_optimal_fontsize(ax=ax, labels=labels, max_x_dist=max_x_dist)
        return round(opt_fs, 2)

    def add_tmd_jmd_seq(self, ax=None, jmd_n_seq=None, tmd_seq=None, jmd_c_seq=None,
                        tmd_color="mediumspringgreen", jmd_color="blue",
                        tmd_seq_color="black", jmd_seq_color="white",
                        add_xticks_pos=False, heatmap=True,
                        x_shift=0, xtick_size=11, seq_size=None,
                        fontsize_tmd_jmd=None, weight_tmd_jmd="normal"):
        """Add sequences and corresponding x-ticks for TMD and JMD regions."""
        seq_size = _add_part_seq(ax=ax, jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq, x_shift=x_shift,
                               seq_size=seq_size, tmd_color=tmd_color, jmd_color=jmd_color,
                               tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
        fontsize_tmd_jmd = seq_size if fontsize_tmd_jmd is None else fontsize_tmd_jmd
        # Set second axis (with ticks and part annotations)
        name_tmd = ut.options["name_tmd"]
        name_jmd_n = ut.options["name_jmd_n"]
        name_jmd_c = ut.options["name_jmd_c"]
        jmd_n_end, tmd_end, jmd_c_end = self._get_ends(x_shift=-1)
        jmd_n_middle, tmd_middle, jmd_c_middle = self._get_middles(x_shift=-0.5)
        # Get x-ticks and x-tick labels
        if not add_xticks_pos:
            xticks = [jmd_n_middle, tmd_middle, jmd_c_middle]
            xtick_labels = [name_jmd_n, name_tmd, name_jmd_c]
        else:
            xticks = [0, jmd_n_middle, jmd_n_end, tmd_middle, tmd_end, jmd_c_middle, jmd_c_end]
            xtick_labels = [self.start, name_jmd_n, jmd_n_end + self.start, name_tmd, tmd_end + self.start, name_jmd_c,
                            jmd_c_end + self.start]
        # Adjust x-ticks and x-tick labels
        exists_jmd_n = len(jmd_n_seq) > 0
        exists_jmd_c = len(jmd_c_seq) > 0
        xticks, xtick_labels = _adjust_xticks_labels(xticks=xticks, xtick_labels=xtick_labels,
                                                     add_xtick_pos=add_xticks_pos,
                                                     exists_jmd_n=exists_jmd_n, exists_jmd_c=exists_jmd_c)
        # Set x-ticks and x-tick labels
        if fontsize_tmd_jmd > 0:
            ax2 = _get_new_axis(ax=ax)
            ax2.set_xlim(ax.get_xlim())
            _add_part_seq_second_ticks(ax2=ax2, xticks=xticks, xtick_labels=xtick_labels,
                                       x_shift=x_shift, seq_size=seq_size,xtick_size=xtick_size,
                                       fontsize_tmd_jmd=fontsize_tmd_jmd, weight_tmd_jmd=weight_tmd_jmd,
                                       heatmap=heatmap)
        return ax

    # Add TMD-JMD elements
    def add_tmd_jmd_bar(self, ax=None, x_shift=0, jmd_color="blue", tmd_color="mediumspringgreen"):
        """Add colored bars to indicate TMD and JMD regions."""
        jmd_n_start, tmd_start, jmd_c_start = self._get_starts(x_shift=x_shift)
        _add_part_bar(ax=ax, start=jmd_n_start, len_part=self.jmd_n_len, color=jmd_color)
        _add_part_bar(ax=ax, start=tmd_start, len_part=self.tmd_len, color=tmd_color)
        _add_part_bar(ax=ax, start=jmd_c_start, len_part=self.jmd_c_len, color=jmd_color)

    def add_tmd_jmd_text(self, ax=None, x_shift=0, fontsize_tmd_jmd=None, weight_tmd_jmd="normal"):
        """Add text labels for TMD and JMD regions."""
        name_tmd = ut.options["name_tmd"]
        name_jmd_n = ut.options["name_jmd_n"]
        name_jmd_c = ut.options["name_jmd_c"]
        jmd_n_start, tmd_start, jmd_c_start = self._get_starts(x_shift=x_shift)
        exists_jmd_n = self.jmd_n_len > 0
        exists_jmd_c = self.jmd_c_len > 0
        if fontsize_tmd_jmd is None or fontsize_tmd_jmd > 0:
            args = dict(ax=ax, fontsize=fontsize_tmd_jmd, fontweight=weight_tmd_jmd)
            _add_part_text(start=tmd_start, len_part=self.tmd_len, text=name_tmd, **args)
            if exists_jmd_n:
                _add_part_text(start=jmd_n_start, text=name_jmd_n, len_part=self.jmd_n_len, **args)
            if exists_jmd_c:
                _add_part_text(start=jmd_c_start, text=name_jmd_c, len_part=self.jmd_c_len, **args)

    def add_tmd_jmd_xticks(self, ax=None, x_shift=0, xtick_size=11.0, xtick_width=2.0, xtick_length=5.0):
        """Adjust x-ticks for TMD and JMD regions."""
        # Remove the xticks and return early
        if xtick_size == 0:
            ax.set_xticks([])
            ax.set_xticklabels([])
            return
        # Adjust tick tick_length based on figure size
        width, height = plt.gcf().get_size_inches()
        xtick_length += height if xtick_length != 0 else 0
        # Adjust for start counting at position 0
        jmd_n_end, tmd_end, jmd_c_end = self._get_ends(x_shift=-1)
        exists_jmd_n = self.jmd_n_len > 0
        exists_jmd_c = self.jmd_c_len > 0
        xticks = [0]
        if exists_jmd_n:
            xticks.append(jmd_n_end)
        xticks.append(tmd_end)
        if exists_jmd_c:
            xticks.append(jmd_c_end)
        ax.set_xticks([x + x_shift for x in xticks])
        ax.set_xticklabels([x + self.start for x in xticks], size=xtick_size, rotation=0)
        ax.tick_params(axis="x", length=xtick_length, color="black", width=xtick_width, bottom=True)


    def highlight_tmd_area(self, ax=None, x_shift=0, tmd_color="mediumspringgreen", alpha=0.2):
        """Highlight the TMD area in the plot."""
        jmd_n_start, tmd_start, jmd_c_start = self._get_starts(x_shift=x_shift)
        y_min, y_max = plt.ylim()
        height = abs(y_min) + y_max
        rect = mpl.patches.Rectangle((tmd_start, y_min), width=self.tmd_len, height=height, linewidth=0,
                                     color=tmd_color, zorder=0.1, clip_on=True, alpha=alpha)
        ax.add_patch(rect)

    # Add x-ticks
    def get_xticks_with_labels(self, step=5):
        """Generate x-ticks and their labels for the plot."""
        second_pos = int((self.start+step)/step)*step
        xticks_middle = list(range(second_pos, self.stop, step))
        xticks_labels = [self.start] + xticks_middle + [self.stop]
        xticks = [x-self.start for x in xticks_labels]
        return xticks, xticks_labels


    def add_xticks(self, ax=None, x_shift=0.0, xticks_position="bottom",
                   xtick_size=11.0, xtick_width=2.0, xtick_length=4.0):
        """Add x-ticks to the plot."""
        xticks, xticks_labels = self.get_xticks_with_labels(step=5)
        ax.set_xticks([x + x_shift for x in xticks])
        ax.set_xticklabels(xticks_labels, rotation=0, size=xtick_size)
        ax.tick_params(axis="x", color="black", length=xtick_length, width=xtick_width)
        ax.xaxis.set_ticks_position(xticks_position)


from typing import Union, NewType, Sequence
import numpy as np
import pandas as pd

# Define the basic numeric type
NumericType = Union[int, float]

# Define the 1D and 2D types using Union
ArrayLike1DUnion = Union[Sequence[NumericType], np.ndarray, pd.Series]
ArrayLike2DUnion = Union[Sequence[Sequence[NumericType]], np.ndarray, pd.DataFrame]

# A 1D array-like object. Can be a sequence (e.g., list or tuple) of ints/floats, numpy ndarray, or pandas Series.
ArrayLike1D = NewType("ArrayLike1D", ArrayLike1DUnion)

# A 2D array-like object. Can be a sequence of sequence of ints/floats, numpy ndarray, or pandas DataFrame.
ArrayLike2D = NewType("ArrayLike2D", ArrayLike2DUnion)

# Numeric type lists
VALID_INT_TYPES = (
    int,
    np.int_, np.intc, np.intp, np.integer,  # np.integer covers all standard integer types
    np.int8, np.int16, np.int32, np.int64,
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.longlong  # Equivalent to np.int64 on most platforms
)

VALID_FLOAT_TYPES = (
    float,
    np.float_, np.float16, np.float32, np.float64, np.longdouble  # np.longdouble is the highest precision float available
)

VALID_INT_FLOAT_TYPES = VALID_INT_TYPES + VALID_FLOAT_TYPES



"""
This is a script for the backend PlotElements utility class for the CPPPlot class.
"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import aaanalysis.utils as ut

# I Helper Functions
DICT_LEGEND_CAT = dict(loc=9,
                       weight_title="normal",
                       labelspacing=0.05,
                       handletextpad=0.2)


def _get_colors_for_col_cat(labels=None, dict_color=None, df_feat=None, col_cat=None):
    """Extend color dict for given category column"""
    dict_cat = dict(zip(df_feat[col_cat], df_feat[ut.COL_CAT]))
    colors = [dict_color[dict_cat[c]] for c in labels]
    return colors


# II Main Functions
class PlotElements:
    """Utility class for plot element configurations and enhancements."""

    # Set plot elements
    @staticmethod
    def set_figsize(ax=None, figsize=None, force_set=False):
        """Set the figure size. Optionally create a figure and an axes object if none exists."""
        # DEV: figsize is not used as argument in seaborn (but in pandas)
        if ax:
            fig = ax.figure
        elif not plt.get_fignums():
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        # Set figure size if force_set is True or if there are no axes in the figure
        if force_set or len(fig.get_axes()) == 0:
            fig.set_size_inches(figsize, forward=True)
        return fig, ax

    @staticmethod
    def set_title_(title=None, title_kws=None):
        """Set the title of the plot."""
        if title_kws is None:
            title_kws = {}
        plt.title(title, **title_kws)

    @staticmethod
    def set_y_ticks(ax=None, fs=None):
        """Set custom y-tick labels with the specified font size."""
        tick_positions = ax.get_yticks()
        ax.yaxis.set_major_locator(mpl.ticker.FixedLocator(tick_positions))
        tick_labels = [round(float(i), 1) for i in tick_positions]
        ax.set_yticklabels(tick_labels, size=fs)

    @staticmethod
    def set_x_ticks(ax=None, fs=None):
        """Set custom x-tick labels with the specified font size."""
        tick_positions = ax.get_xticks()
        ax.xaxis.set_major_locator(mpl.ticker.FixedLocator(tick_positions))
        tick_labels = [round(float(i), 1) for i in tick_positions]
        ax.set_xticklabels(tick_labels, size=fs)

    # Scale classification elements
    @staticmethod
    def add_subcat_bars(ax=None, df_pos=None, df_feat=None, col_cat=None, dict_color=None,
                        bar_width=0.3, bar_spacing=0.15):
        """Add left colored sidebar to indicate category grouping"""
        labels = list(df_pos.index)
        colors = _get_colors_for_col_cat(labels=labels, df_feat=df_feat, col_cat=col_cat, dict_color=dict_color)
        ut.plot_add_bars(ax=ax, labels=labels, colors=colors,
                         bar_width=bar_width, bar_spacing=bar_spacing, label_spacing_factor=2)


    @staticmethod
    def adjust_cat_legend_kws(legend_kws=None, n_cat=None,
                              legend_xy=None, fontsize_labels=None):
        """Optimize legend position and appearance based on the number of categories and provided keywords."""
        n_cols = 2 if legend_kws is None else legend_kws.get("n_cols", 2)
        n_rows = np.floor(n_cat / n_cols)

        # Create legend position [center (x), top (y)]
        legend_xy_default = (-0.1, -0.01)
        _legend_xy = ut.adjust_tuple_elements(tuple_in=legend_xy,
                                              tuple_default=legend_xy_default)
        x, y = _legend_xy
        str_space = "\n" * int((6-n_rows))
        title = f"{str_space}{ut.LABEL_SCALE_CAT}"
        # Prepare legend keywords
        _legend_kws = dict(fontsize=fontsize_labels,
                           fontsize_title=fontsize_labels,
                           n_cols=n_cols,
                           title=title,
                           x=x, y=y,
                           **DICT_LEGEND_CAT)
        if legend_kws is not None:
            _legend_kws.update(legend_kws)
        return _legend_kws


    # Colorbar elements
    @staticmethod
    def adjust_cbar_kws(fig=None, cbar_kws=None, cbar_xywh=None,
                        fontsize_labels=None, label=None):
        """Set colorbar position, appearance, and label with default or provided keywords."""
        width, height = plt.gcf().get_size_inches()
        bar_height = 0.15/height
        bar_bottom = 0.06/height

        # Create cbar positions: [left (x), bottom (y), width, height]
        cbar_xywh_default = (0.5, bar_bottom, 0.2, bar_height)
        _cbar_xywh = ut.adjust_tuple_elements(tuple_in=cbar_xywh,
                                              tuple_default=cbar_xywh_default)

        # Create colorbar axes
        cbar_ax = fig.add_axes(_cbar_xywh)

        # Prepare colorbar keywords
        _cbar_kws = dict(ticksize=fontsize_labels, label=label)
        if cbar_kws is not None:
            _cbar_kws.update(cbar_kws)
        return _cbar_kws, cbar_ax



"""
This is a script for data checking utility functions.
"""
import pandas as pd
import numpy as np
from sklearn.utils import check_array

from ._utils import add_str
from .utils_types import VALID_INT_TYPES, VALID_INT_FLOAT_TYPES
import aaanalysis._utils.check_type as check_type


# Helper functions
def _convert_2d(val=None, name=None, str_add=None):
    """
    Convert array-like data to 2D array. Handles lists of arrays, lists of lists, and 1D lists.
    """
    str_error = add_str(str_error=f"'{name}' should be a 2D list or 2D array with rows having the same number of columns.",
                        str_add=str_add)
    if isinstance(val, list):
        # Check if List with arrays and return if yes
        if all(isinstance(i, np.ndarray) for i in val):
            try:
                val = np.asarray(val)
            except ValueError:
                raise ValueError(str_error)
        # Convert 1D list to 2D list
        elif all(not isinstance(i, list) for i in val):
            try:
                val = np.asarray([val])
            except ValueError:
                raise ValueError(str_error)
        # For nested lists, ensure they are 2D (list of lists with equal lengths)
        else:
            try:
                val = np.array(val)  # Convert nested list to numpy array
                if val.ndim != 2:
                    raise ValueError
            except ValueError:
                raise ValueError(str_error)
    elif hasattr(val, 'ndim') and val.ndim == 1:
        try:
            val = np.asarray([val])
        except ValueError:
            raise ValueError(str_error)
    return val


# Check array like
def check_array_like(name=None, val=None, dtype=None, ensure_2d=False, allow_nan=False, convert_2d=False,
                     accept_none=False, expected_dim=None, str_add=None):
    """Check if the provided value is array-like and matches the specified dtype."""
    if val is None:
        if accept_none:
            return None  # Skip tests
        else:
            raise ValueError(f"'{name}' should not be None.")
    # Extend dtype to handle a list of dtypes including bool
    dtype = check_type.check_str(name="dtype", val=dtype, accept_none=True)
    valid_dtypes = ["numeric", "int", "float", "bool", None]
    if dtype not in valid_dtypes:
        str_error = add_str(str_error=f"'dtype' should be one of the following: {valid_dtypes}", str_add=str_add)
        raise ValueError(str_error)
    dict_expected_dtype = {"numeric": "numeric", "int": "int64", "float": "float64", "bool": "bool"}
    expected_dtype = dict_expected_dtype[dtype] if dtype is not None else None
    # Specific check for boolean arrays
    if dtype == "bool":
        flattened_val = np.array(val).flatten()
        if not all(isinstance(item, (bool, np.bool_)) for item in flattened_val):
            str_error = add_str(str_error=f"All elements in '{name}' must be of type 'bool' (either Python native or NumPy bool).",
                                str_add=str_add)
            raise ValueError(str_error)
    # Convert a 1D list or array to a 2D array if needed
    if convert_2d:
        val = _convert_2d(val=val, name=name, str_add=str_add)
    # Utilize Scikit-learn's check_array for robust checking
    try:
        # Convert list to array
        val = check_array(val, dtype=expected_dtype, ensure_2d=ensure_2d, force_all_finite=not allow_nan)
    except Exception as e:
        dtype = "any type" if dtype is None else dtype
        raise ValueError(f"'{name}' should be array-like with '{dtype}' values."
                         f"\nScikit message:\n\t{e}")
    # Check dimensions if specified
    if expected_dim is not None and len(val.shape) != expected_dim:
        str_error = add_str(str_error=f"'{name}' should have {expected_dim} dimensions, but has {len(val.shape)}.",
                            str_add=str_add)
        raise ValueError(str_error)
    return val


# Check feature matrix and labels
def check_X(X, X_name="X", min_n_samples=3, min_n_features=2, min_n_unique_features=None,
            ensure_2d=True, allow_nan=False, accept_none=False, str_add=None):
    """Check the feature matrix X is valid."""
    if X is None:
        if not accept_none:
            raise ValueError(f"'{X_name}' should not be None")
        else:
            return None
    X = check_array_like(name="X", val=X, dtype="float", ensure_2d=ensure_2d, allow_nan=allow_nan)
    if np.isinf(X).any():
        str_error = add_str(str_error=f"'X' should not contain infinite values",
                            str_add=str_add)
        raise ValueError(str_error)
    n_samples, n_features = X.shape
    if n_samples < min_n_samples:
        raise ValueError(f"n_samples ({n_samples} in 'X') should be >= {min_n_samples}."
                         f"\nX = {X}")
    if n_features < min_n_features:
        raise ValueError(f"n_features ({n_features} in 'X') should be >= {min_n_features}."
                         f"\nX = {X}")
    if min_n_unique_features is not None:
        n_unique_features = sum([len(set(X[:, col])) > 1 for col in range(n_features)])
        if n_unique_features < min_n_unique_features:
            str_error = add_str(str_error=f"'n_unique_features' ({n_unique_features}) should be >= {min_n_unique_features}",
                                str_add=str_add)
            raise ValueError(str_error)
    return X


def check_X_unique_samples(X, min_n_unique_samples=3, str_add=None):
    """Check if the matrix X has a sufficient number of unique samples."""
    n_unique_samples = len(set(map(tuple, X)))
    if n_unique_samples == 1:
        str_error = add_str(str_error="Feature matrix 'X' should not have all identical samples.",
                            str_add=str_add)
        raise ValueError(str_error)
    if n_unique_samples < min_n_unique_samples:
        raise ValueError(f"n_unique_samples ({n_unique_samples}) should be >= {min_n_unique_samples}."
                         f"\nX = {X}")
    return X


def check_labels(labels=None, name="labels", vals_requiered=None, len_requiered=None, allow_other_vals=True,
                 n_per_group_requiered=None, accept_float=False, str_add=None):
    """Check the provided labels against various criteria like type, required values, and length."""
    if labels is None:
        raise ValueError(f"'{name}' should not be None.")
    labels = check_type.check_list_like(name=name, val=labels)
    # Convert labels to a numpy array if it's not already
    labels = np.asarray(labels)
    # Ensure labels is at least 1-dimensional
    if labels.ndim == 0:
        labels = np.array([labels.item()])  # Convert 0-d array to 1-d array
    unique_labels = set(labels)
    if len(unique_labels) == 1:
        str_error = add_str(str_error=f"'{name}' should contain more than one different value ({unique_labels}).",
                            str_add=str_add)
        raise ValueError(str_error)
    valid_types = VALID_INT_TYPES if not accept_float else VALID_INT_FLOAT_TYPES
    wrong_types = [l for l in unique_labels if not isinstance(l, valid_types)]
    if wrong_types:
        str_error = add_str(str_error=f"Labels in '{name}' should be type int, but contain: {set(map(type, wrong_types))}",
                            str_add=str_add)
        raise ValueError(str_error)
    if vals_requiered is not None:
        missing_vals = [x for x in vals_requiered if x not in labels]
        if len(missing_vals) > 0:
            str_error = add_str(str_error=f"'{name}' ({unique_labels}) does not contain requiered values: {missing_vals}",
                                str_add=str_add)
            raise ValueError(str_error)
        if not allow_other_vals:
            wrong_vals = [x for x in labels if x not in vals_requiered]
            if len(wrong_vals) > 0:
                str_error = add_str(str_error=f"'{name}' ({unique_labels}) does contain wrong values: {wrong_vals}",
                                    str_add=str_add)
                raise ValueError(str_error)
    if len_requiered is not None and len(labels) != len_requiered:
        str_error = add_str(str_error=f"'{name}' (n={len(labels)}) should contain {len_requiered} values.",
                            str_add=str_add)
        raise ValueError(str_error)
    # Check for minimum length per group
    if n_per_group_requiered is not None:
        label_counts = {label: np.sum(labels == label) for label in unique_labels}
        underrepresented_labels = {label: count for label, count in label_counts.items() if
                                   count < n_per_group_requiered}
        if underrepresented_labels:
            str_error = add_str(str_error=f"Each label should have at least {n_per_group_requiered} occurrences. "
                                          f"Underrepresented labels: {underrepresented_labels}",
                                str_add=str_add)
            raise ValueError(str_error)
    return labels


def check_match_X_labels(X=None, X_name="X", labels=None, labels_name="labels", check_variability=False,
                         str_add=None):
    """Check if the number of samples in X matches the number of labels."""
    n_samples, n_features = X.shape
    if n_samples != len(labels):
        str_error = add_str(str_error=f"n_samples does not match for '{X_name}' ({len(X)}) and '{labels_name}' ({len(labels)}).",
                            str_add=str_add)
        raise ValueError(str_error)
    if check_variability:
        unique_labels = np.unique(labels)
        for label in unique_labels:
            group_X = X[labels == label]
            if not np.all(np.var(group_X, axis=0) != 0):
                str_error = add_str(str_error=f"Variance in 'X' for label '{label}' from '{labels_name}' is too low.",
                                    str_add=str_add)
                raise ValueError(str_error)


def check_match_X_list_labels(X=None, list_labels=None, check_variability=False, vals_requiered=None, str_add=None):
    """Check if each label set is matching with X"""
    for i, labels in enumerate(list_labels):
        check_labels(labels=labels, vals_requiered=vals_requiered)
        check_match_X_labels(X=X, labels=labels, labels_name=f"list_labels (set {i+1})",
                             check_variability=check_variability, str_add=str_add)


def check_match_list_labels_names_datasets(list_labels=None, names_datasets=None, str_add=None):
    """Check if length of list_labels and names match"""
    if names_datasets is None:
        return None # Skip check
    if len(list_labels) != len(names_datasets):
        str_error = add_str(str_error=f"Length of 'list_labels' ({len(list_labels)}) and 'names_datasets'"
                                      f" ({len(names_datasets)} does not match)",
                            str_add=str_add)
        raise ValueError(str_error)



# Check sets
def check_superset_subset(subset=None, superset=None, name_subset=None, name_superset=None, str_add=None):
    """Check if all elements of the subset are contained in the superset."""
    wrong_elements = [x for x in subset if x not in superset]
    if len(wrong_elements) != 0:
        str_error = add_str(str_error=f"'{name_superset}' does not contain the following elements of '{name_subset}': {wrong_elements}",
                            str_add=str_add)
        raise ValueError(str_error)


# df checking functions
def check_df(name="df", df=None, accept_none=False, accept_nan=True, check_all_positive=False,
             cols_requiered=None, cols_forbidden=None, cols_nan_check=None, str_add=None):
    """Check if the provided DataFrame meets various criteria such as NaN values, required/forbidden columns, etc."""
    # Check DataFrame and values
    if df is None:
        if not accept_none:
            raise ValueError(f"'{name}' should not be None")
        else:
            return None
    if not isinstance(df, pd.DataFrame):
        str_error = add_str(str_error= f"'{name}' ({type(df)}) should be DataFrame",
                            str_add=str_add)
        raise ValueError(str_error)
    if not accept_nan and df.isna().any().any():
        str_error = add_str(str_error=f"'{name}' contains NaN values, which are not allowed",
                            str_add=str_add)
        raise ValueError(str_error)
    if check_all_positive:
        numeric_df = df.select_dtypes(include=['float', 'int'])
        if numeric_df.min().min() <= 0:
            str_error = add_str(str_error=f"'{name}' should not contain non-positive values.",
                                str_add=str_add)
            raise ValueError(str_error)
    # Check columns
    args = dict(accept_str=True, accept_none=True, str_add=str_add)
    cols_requiered = check_type.check_list_like(name='cols_requiered', val=cols_requiered, **args)
    cols_forbidden = check_type.check_list_like(name='cols_forbidden', val=cols_forbidden, **args)
    cols_nan_check = check_type.check_list_like(name='cols_nan_check', val=cols_nan_check, **args)
    if cols_requiered is not None:
        missing_cols = [col for col in cols_requiered if col not in df.columns]
        if len(missing_cols) > 0:
            str_error = add_str(str_error=f"'{name}' is missing required columns: {missing_cols}",
                                str_add=str_add)
            raise ValueError(str_error)
    if cols_forbidden is not None:
        forbidden_cols = [col for col in cols_forbidden if col in df.columns]
        if len(forbidden_cols) > 0:
            str_error = add_str(str_error=f"'{name}' is contains forbidden columns: {forbidden_cols}",
                                str_add=str_add)
            raise ValueError(str_error)
    if cols_nan_check is not None:
        if df[cols_nan_check].isna().sum().sum() > 0:
            str_error = add_str(str_error=f"NaN values are not allowed in '{cols_nan_check}'.",
                                str_add=str_add)
            raise ValueError(str_error)



"""
This is a script for the frontend of the CPPPlot class.
"""
from typing import Optional, Dict, Union, List, Tuple, Type, Literal
import pandas as pd
import matplotlib.pyplot as plt
import warnings

import aaanalysis.utils as ut

from ._backend.check_feature import (check_split_kws,
                                     check_parts_len,
                                     check_match_features_seq_parts,
                                     check_df_parts,
                                     check_match_df_parts_features,
                                     check_match_df_parts_list_parts,
                                     check_df_scales,
                                     check_match_df_scales_features,
                                     check_df_cat,
                                     check_match_df_cat_features,
                                     check_match_df_parts_df_scales,
                                     check_match_df_seq_jmd_len,
                                     check_match_df_scales_df_cat)
from ._backend.check_cpp_plot import (check_args_xtick,
                                      check_args_ytick,
                                      check_part_color,
                                      check_seq_color,
                                      check_match_dict_color_df)

from ._backend.cpp.cpp_plot_eval import plot_eval
from ._backend.cpp.cpp_plot_feature import plot_feature
from ._backend.cpp.cpp_plot_ranking import plot_ranking
from ._backend.cpp.cpp_plot_profile import plot_profile
from ._backend.cpp.cpp_plot_heatmap import plot_heatmap
from ._backend.cpp.cpp_plot_feature_map import plot_feature_map
from ._backend.cpp.cpp_plot_update_seq_size import get_tmd_jmd_seq, update_seq_size_, update_tmd_jmd_labels


# I Helper Functions
# Checks for eval plot
def check_match_dict_color_list_cat(dict_color=None, list_cat=None):
    """Check if all categories from list_cat are in dict_color"""
    if list_cat is None:
        return dict_color # Skip test
    # Check list_cat for duplicates
    if len(list_cat) != len(set(list_cat)):
        raise ValueError(f"'list_cat' {list_cat} should not contain duplicates.")

    missing_cat = [x for x in list_cat if x not in dict_color]
    if len(missing_cat) > 0:
        raise ValueError(f"Following categories from 'list_cat' miss a color in 'dict_color': {missing_cat}")
    return {cat: dict_color[cat] for cat in list_cat}


def check_df_eval(df_eval):
    """Check if columns in df_eval have valid values"""
    if len(df_eval) <= 1:
        raise ValueError("'df_eval' should contain at least two features sets")
    # Check quality measures
    list_n_feat = df_eval[ut.COL_N_FEAT].values[0]
    range_abs_auc = df_eval[ut.COL_RANGE_ABS_AUC].values[0]
    avg_mean_dif = df_eval[ut.COL_RANGE_ABS_AUC].values[0]
    n_clusters = df_eval[ut.COL_N_CLUST].values[0]
    avg_n_feat_per_clust = df_eval[ut.COL_AVG_N_FEAT_PER_CLUST].values[0]
    std_n_feat_per_clust = df_eval[ut.COL_STD_N_FEAT_PER_CLUST].values[0]
    ut.check_number_val(name=f"{ut.COL_N_FEAT}: n_features", val=list_n_feat[0], just_int=True)
    ut.check_list_like(name=f"{ut.COL_N_FEAT}: list_n_feat", val=list_n_feat[1], accept_str=False)
    ut.check_list_like(name=ut.COL_RANGE_ABS_AUC, val=range_abs_auc, accept_str=False)
    ut.check_number_val(name=f"{ut.COL_AVG_MEAN_DIF} (pos)", val=avg_mean_dif[0], just_int=False)
    ut.check_number_val(name=f"{ut.COL_AVG_MEAN_DIF} (neg)", val=avg_mean_dif[1], just_int=False)
    ut.check_number_range(name=ut.COL_N_CLUST, val=n_clusters, min_val=1, just_int=True)
    ut.check_number_range(name=ut.COL_AVG_N_FEAT_PER_CLUST, val=avg_n_feat_per_clust, just_int=False)
    ut.check_number_range(name=ut.COL_STD_N_FEAT_PER_CLUST, val=std_n_feat_per_clust, just_int=False)


def check_match_df_eval_list_cat(df_eval=None, list_cat=None):
    """Check if number of features per category in df_eval and list_cat match"""
    names = df_eval[ut.COL_NAME].to_list()
    list_n_feat_sets = [x[1] for x in df_eval[ut.COL_N_FEAT]]
    for list_n_feat, name in zip(list_n_feat_sets, names):
        if len(list_n_feat) != len(list_cat):
            raise ValueError(f"Number of features per category in '{name}' does not match with 'list_cat' {list_cat}")


# Checks for feature plot
def check_match_df_seq_names_to_show(df_seq=None, names_to_show=None):
    """Check if """
    if names_to_show is None:
        return # Skip check
    if ut.COL_NAME not in df_seq:
        raise ValueError(f"'df_seq' must contain '{ut.COL_NAME}' column if 'names_to_show' ({names_to_show}) is not none")
    list_names = df_seq[ut.COL_NAME].to_list()
    missing_names = [x for x in names_to_show if x not in list_names]
    if len(missing_names) > 0:
        raise ValueError(f"Following names from 'names_to_show' are not in '{ut.COL_NAME}' "
                         f"column of 'df_seq': {missing_names}")


# Checks for main CPP plots (ranking, profile, maps)
def check_cmap_for_heatmap(cmap=None):
    """Check if cmap is valid or 'SHAP'"""
    if cmap in [ut.STR_CMAP_SHAP, ut.STR_CMAP_CPP]:
        return None # Skip test
    ut.check_cmap(name="cmap", val=cmap, accept_none=True)


def check_col_dif(col_dif=None, shap_plot=False):
    """Check if col_dif is string and set default"""
    ut.check_str(name="col_dif", val=col_dif, accept_none=False)
    if col_dif is None:
        col_dif = ut.COL_MEAN_DIF
    if not shap_plot:
        if col_dif != ut.COL_MEAN_DIF:
            raise ValueError(f"'col_dif' ('{col_dif}') must be '{ut.COL_MEAN_DIF}'")
    else:
        if ut.COL_MEAN_DIF not in col_dif:
            raise ValueError(f"If 'shap_plot=True', 'col_dif' ('{col_dif}') must follow '{ut.COL_MEAN_DIF}_'name''")


def check_col_imp(col_imp=None, shap_plot=False):
    """Check if col_imp is string and set default"""
    ut.check_str(name="col_imp", val=col_imp, accept_none=True)
    if col_imp is None:
        col_imp = ut.COL_FEAT_IMPACT if shap_plot else ut.COL_FEAT_IMPORT
    if not shap_plot:
        if ut.COL_FEAT_IMPORT not in col_imp:
            raise ValueError(f"'col_imp' ('{col_imp}') must be '{ut.COL_FEAT_IMPORT} or follow '{ut.COL_FEAT_IMPORT}_'name'")
    else:
        if ut.COL_FEAT_IMPACT not in col_imp:
            raise ValueError(f"If 'shap_plot=True', 'col_imp' ('{col_imp}') must follow '{ut.COL_FEAT_IMPACT}_'name''")
    return col_imp


def check_col_cat(col_cat=None):
    """Check if col_cat is valid"""
    list_valid_col_cat = [ut.COL_CAT, ut.COL_SUBCAT, ut.COL_SCALE_NAME]
    ut.check_str(name="col_cat", val=col_cat, str_add=f"Should be one of the following: {list_valid_col_cat}")
    if col_cat not in list_valid_col_cat:
        raise ValueError(f"'col_cat' ({col_cat}) should be one of the following: {list_valid_col_cat}")


def check_col_val(col_val=None, shap_plot=False, sample_mean_dif=False):
    """Check if col_val is valid"""
    list_valid_col_val = [ut.COL_MEAN_DIF, ut.COL_ABS_MEAN_DIF, ut.COL_ABS_AUC, ut.COL_FEAT_IMPORT]
    str_error_shap = (f"If 'shap_plot=True', 'col_val' ('{col_val}') must follow '{ut.COL_FEAT_IMPACT}_'name'' or "
                      f"'{ut.COL_MEAN_DIF}_name'")
    str_add = f"Should be one of the following: {list_valid_col_val}" if not shap_plot else str_error_shap
    ut.check_str(name="col_val", val=col_val, accept_none=False, str_add=str_add)
    if not shap_plot:
        if sample_mean_dif:
            if ut.COL_MEAN_DIF not in col_val and col_val not in list_valid_col_val:
                raise ValueError(f"'col_val' ('{col_val}') must follow {ut.COL_MEAN_DIF}_name' or "
                                 f"should be one of the following: {list_valid_col_val}")
        elif col_val not in list_valid_col_val:
            raise ValueError(f"'col_val' ('{col_val}') should be one of the following: {list_valid_col_val}")
    else:
        if ut.COL_FEAT_IMPACT not in col_val and ut.COL_MEAN_DIF not in col_val:
            raise ValueError(str_error_shap)
    return col_val


def check_match_shap_plot_add_legend_cat(shap_plot=False, add_legend_cat=False):
    """Check if not both are True"""
    if shap_plot and add_legend_cat:
        raise ValueError(f"'shap_plot' ({shap_plot}) and 'add_legend_cat' ({add_legend_cat}) can not be both True.")


def check_imp_tuples(name="imp_th", imp_tuples=None):
    """Check if legend importance thresholds are valid"""
    ut.check_tuple(name=name, val=imp_tuples, n=3,
                   accept_none=False, check_number=True,
                   accept_none_number=False, str_add=f"Or should not contain None values: {imp_tuples}")
    th1, th2, th3 = imp_tuples
    if th1 >= th2 or th2 >= th3:
        raise ValueError(f"'{name}' ({imp_tuples}) should contain 3 numbers in ascending order")
    if th1 <= 0:
        raise ValueError(f"Minium of '{name}' ({imp_tuples[0]}) should be > 0")


# Check update_seq_size
def check_match_ax_seq_len(ax=None, jmd_n_len=10, jmd_c_len=10):
    """Check if ax matches with requiered length"""
    labels = ax.xaxis.get_ticklabels(which="both")
    f = lambda l: l.get_window_extent(ax.figure.canvas.get_renderer())
    tick_positions = [f(l).x0 for l in labels]
    _, sorted_labels = zip(*sorted(zip(tick_positions, labels), key=lambda t: t[0]))
    tmd_jmd_seq = "".join([x.get_text() for x in sorted_labels])
    min_len = jmd_n_len + 1 + jmd_c_len
    if len(tmd_jmd_seq) <= min_len:
        raise ValueError(f"TMD-JMD sequence from 'ax' is shorter than minimum ({min_len}; jmd_n: {jmd_n_len}, tmd:>1, jmd_c: {jmd_c_len}."
                         f"\n Sequence (len: {len(tmd_jmd_seq)}) retrieved from 'ax' is: '{tmd_jmd_seq}'")


# II Main Functions
class CPPPlot:
    """
    Plotting class for :class:`CPP` (Comparative Physicochemical Profiling) results [Breimann24c]_.

    This class supports multiple plot types for group or sample-level analysis, including ranking plots,
    profiles, heatmaps, and feature maps.

    """
    def __init__(self,
                 df_scales: Optional[pd.DataFrame] = None,
                 df_cat: Optional[pd.DataFrame] = None,
                 jmd_n_len: int = 10,
                 jmd_c_len: int = 10,
                 accept_gaps: bool = False,
                 verbose: bool = True
                 ):
        """
        Parameters
        ----------
        df_scales : pd.DataFrame, shape (n_letters, n_scales), optional
            DataFrame of scales with letters typically representing amino acids. Default from :meth:`load_scales`
            unless specified in ``options['df_scales']``.
        df_cat : pd.DataFrame, shape (n_scales, n_scales_info), optional
            DataFrame of categories for physicochemical scales. Must contain all scales from ``df_scales``.
            Default from :meth:`load_scales` with ``name='scales_cat'``, unless specified in ``options['df_cat']``.
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len: int, default=10
            Length of JMD-C (>=0).
        accept_gaps: bool, default=False
            Whether to accept missing values by enabling omitting for computations (if ``True``).
        verbose: bool, default=True
            If ``True``, verbose outputs are enabled.

        Notes
        -----
        Several methods provide the ``shap_plot`` parameter, which allows to specify whether a plot visualizes
        the results of the group-level CPP analysis or the sample-level CPP-SHAP analysis (if ``shap_plot=True``).

        - **CPP Analysis**: Group-level analysis of the most discriminant features between a test and a reference group.
          The overall results are visualized by the :meth:`CPPPlot.feature_map`, revealing the characteristic
          physicochemical signature of the test group.
        - **CPP-SHAP Analysis**: Sample-level analysis of the CPP feature impact with single-residue resolution.

        The methods showing the CPP and CPP-SHAP analysis results are as follows:

        - :meth:`CPPPlot.ranking`: the 'CPP/-SHAP ranking plot' shows the top n ranked features, their feature value differences,
          and feature importance/impact.
        - :meth:`CPPPlot.profile`: the 'CPP/-SHAP profile' shows the cumulative feature importance/impact per residue position.
        - :meth:`CPPPlot.heatmap`: the 'CPP/-SHAP heatmap' shows the feature value difference or feature impact per
          residue position (x-axis) and scale subcategory (y-axis).

        See Also
        --------
        * :class:`CPP`: the respective computation class for the **CPP Analysis**.
        * :class:`ShapExplainer`: the class combining CPP with the `SHAP <https://shap.readthedocs.io/en/latest/index.html>`_
          explainable Artificial Intelligence (AI) framework.
        * `Anatomy of a figure <https://matplotlib.org/stable/gallery/showcase/anatomy.html>`_ matplotlib guide on
          figure elements.

        Examples
        --------
        .. include:: examples/cpp_plot.rst
        """
        # Load defaults
        if df_scales is None:
            df_scales = ut.load_default_scales()
        if df_cat is None:
            df_cat = ut.load_default_scales(scale_cat=True)
        # Check input
        verbose = ut.check_verbose(verbose)
        check_df_scales(df_scales=df_scales)
        check_df_cat(df_cat=df_cat)
        check_parts_len(jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, accept_none_tmd_len=True)
        ut.check_bool(name="accept_gaps", val=accept_gaps)
        df_scales, df_cat = check_match_df_scales_df_cat(df_cat=df_cat, df_scales=df_scales, verbose=verbose)
        # General settings
        self._verbose = verbose
        self._accept_gaps = accept_gaps
        # Set data parameters
        self._df_cat = df_cat
        self._df_scales = df_scales
        # Set consistent length of JMD-N and JMD-C
        self._jmd_n_len = jmd_n_len
        self._jmd_c_len = jmd_c_len

    @staticmethod
    def eval(df_eval: pd.DataFrame = None,
             figsize: Tuple[int or float, int or float] = (6, 4),
             dict_xlims: Optional[Union[None, dict]] = None,
             legend: bool = True,
             legend_y: float = -0.3,
             dict_color: Optional[dict] = None,
             list_cat: Optional[List[str]] = None,
             ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot evaluation output of CPP comparing multiple sets of identified feature sets.

        Evaluation measures are categorized into two groups:

        * **Discriminative Power** measures ('range_ABS_AUC' and 'avg_MEAN_DIF'), which
          assess the effectiveness of the feature set in distinguishing between the test and reference datasets.
        * **Redundancy** measures ('n_clusters', 'avg_n_feat_per_clust', and 'std_n_feat_per_clust'), which
          evaluate the internal redundancy of a feature set using Pearson correlation-based clustering.

        Parameters
        ----------
        df_eval : pd.DataFrame, shape (n_feature_sets, n_metrics)
            DataFrame with evaluation measures for sets of identified features. Each `row` corresponds to a specific
            feature set. Requiered 'columns' are:

            - 'name': Name of the feature set.
            - 'n_features': Number of features per scale category given as list.
            - 'range_ABS_AUC': Quintile range of absolute AUC among all features (min, 25%, median, 75%, max).
            - 'avg_MEAN_DIF': Two mean differences averaged across all features (for positive and negative 'mean_dif')
            - 'n_clusters': Optimal number of clusters.
            - 'avg_n_feat_per_clust': Average number of features per cluster.
            - 'std_n_feat_per_clust': Standard deviation of feature number per cluster

        figsize : tuple, default=(6, 4)
            Figure dimensions (width, height) in inches.
        dict_xlims : dict, optional
            A dictionary containing x-axis limits for subplots. Keys should be the subplot axis number ({0, 1, 2, 4})
            and values should be tuple specifying (``xmin``, ``xmax``). If ``None``, x-axis limits are auto-scaled.
        legend : bool, default=True
            If ``True``, scale category legend is set under number of features measures.
        legend_y : float, default=-0.3
            Legend position regarding the plot y-axis applied if ``legend=True``.
        dict_color : dict, optional
            Color dictionary of scale categories for legend. Default from :meth:`plot_get_cdict` with ``name='DICT_CAT'``.
        list_cat : list of str, optional
            List of scale categories for which feature numbers are shown.

        Returns
        -------
        fig : plt.Figure
            Figure object for evaluation plot
        axes : array of plt.Axes
            Array of Axes objects, each representing a subplot within the figure.

        Notes
        -----
        * Altering ``figsize`` height could result in unappropriated legend spacing. This can be adjusted by the
          ``legend_y`` parameter together with using the :func:`matplotlib.pyplot.subplots_adjust` function,
          here used with (wspace=0.25, hspace=0, bottom=0.35) parameter settings.

        See Also
        --------
        * :meth:`CPP.eval`: the respective computaton method.
        * :func:`comp_auc_adjusted`.

        Examples
        --------
        .. include:: examples/cpp_plot_eval.rst
        """
        # Check input
        cols_requiered = [ut.COL_NAME, ut.COL_N_FEAT,
                          ut.COL_RANGE_ABS_AUC, ut.COL_AVG_MEAN_DIF,
                          ut.COL_N_CLUST, ut.COL_AVG_N_FEAT_PER_CLUST, ut.COL_STD_N_FEAT_PER_CLUST]
        ut.check_df(name="df_eval", df=df_eval, cols_requiered=cols_requiered, accept_none=False, accept_nan=False)
        check_df_eval(df_eval=df_eval)
        ut.check_figsize(figsize=figsize, accept_none=True)
        ut.check_dict_xlims(dict_xlims=dict_xlims, n_ax=5)
        ut.check_bool(name="legend", val=legend)
        ut.check_number_val(name="legend_y", val=legend_y)
        ut.check_dict_color(name="dict_color", val=dict_color, accept_none=True)
        list_cat = ut.check_list_like(name="list_cat", val=list_cat, accept_none=True, accept_str=True,
                                      check_all_str_or_convertible=True)
        if dict_color is None:
            dict_color = ut.plot_get_cdict_(name="DICT_CAT")
        if list_cat is None:
            list_cat = ut.LIST_CAT
        check_match_df_eval_list_cat(df_eval=df_eval, list_cat=list_cat)
        dict_color = check_match_dict_color_list_cat(dict_color=dict_color, list_cat=list_cat)
        # Plotting
        fig, axes = plot_eval(df_eval=df_eval, figsize=figsize, dict_xlims=dict_xlims,
                              legend=legend, legend_y=legend_y,
                              dict_color=dict_color, list_cat=list_cat)
        return fig, axes


    # Plotting method for single feature
    def feature(self,
                feature: str = None,
                df_seq: pd.DataFrame = None,
                labels: ut.ArrayLike1D = None,
                label_test: int = 1,
                label_ref: int = 0, 
                ax: Optional[plt.Axes] = None,
                figsize: Tuple[Union[int, float], Union[int, float]] = (5.6, 4.8),
                names_to_show: Optional[Union[List[str], str]] = None,
                name_test: str = "TEST",
                name_ref: str = "REF",
                color_test: str = "tab:green",
                color_ref: str = "tab:gray",
                show_seq: bool = False,
                histplot: bool = False,
                fontsize_mean_dif: Union[int, float, None] = 15,
                fontsize_name_test: Union[int, float, None] = 13,
                fontsize_name_ref: Union[int, float, None] = 13,
                fontsize_names_to_show: Union[int, float, None] = 11,
                alpha_hist: int or float = 0.1,
                alpha_dif: int or float = 0.2,
                ) -> plt.Axes:
        """
        Plot distributions of CPP feature values for test and reference datasets highlighting their mean difference.

        Introduced in [Breimann24a]_, a CPP feature is defined as a ``Part-Split-Scale`` combination. For a sample,
        a feature value is computed in three steps:

            1. **Part Selection**: Identify a specific sequence part.
            2. **Part-Splitting**: Divide the selected part into segments, creating a 'Part-Split' combination.
            3. **Scale Value Assignment**: For each amino acid in the 'Part-Split' segment,
               assign its corresponding scale value and calculate the average, which is termed the feature value.

        Parameters
        ----------
        feature : str
            Name of the feature for which test and reference set distributions and difference should be plotted.
        df_seq : pd.DataFrame, shape (n_samples, n_seq_info)
            DataFrame containing an ``entry`` column with unique protein identifiers and sequence information
            in a distinct ``Position-based``, ``Part-based``, ``Sequence-based``, or ``Sequence-TMD-based`` format.
        labels : array-like, shape (n_samples,)
            Class labels for samples in sequence DataFrame (typically, test=1, reference=0).
        label_test : int, default=1,
            Class label of test group in ``labels``.
        label_ref : int, default=0,
            Class label of reference group in ``labels``.
        ax : plt.Axes, optional
            Pre-defined Axes object to plot on. If ``None``, a new Axes object is created.
        figsize : tuple, default=(5.6, 4.8)
            Figure dimensions (width, height) in inches.
        names_to_show : list of str, optional
            Names of specific samples from ``df_seq`` to highlight on plot. 'name' column must be given in ``df_seq``
            if ``names_to_show`` is not ``None``.
        name_test : str, default="TEST"
            Name for the test dataset.
        name_ref : str, default="REF"
            Name for the reference dataset.
        color_test : str, default="tab:green"
            Color for the test dataset.
        color_ref : str, default="tab:gray"
            Color for the reference dataset.
        show_seq : bool, default=False
            If ``True``, show sequence of samples selected via ``names_to_show``.
        histplot : bool, default=False
            If ``True``, plot a histogram. If ``False``, plot a kernel density estimate (KDE) plot.
        fontsize_mean_dif : int or float, default=15
            Font size (>0) for displayed mean difference text.
        fontsize_name_test : int or float, default=13
            Font size (>0) for the name of the test dataset.
        fontsize_name_ref : int or float, default=13
            Font size (>0) for the name of the reference dataset.
        fontsize_names_to_show : int or float, default=11
            Font size (>0) for the names selected via ``names_to_show``.
        alpha_hist : int or float, default=0.1
            The transparency alpha value [0-1] for the histogram distributions.
        alpha_dif : int or float, default=0.2
            The transparency alpha value [0-1] for the mean difference area.

        Returns
        -------
        ax : plt.Axes
            CPP feature plot axes object.

        See Also
        --------
        * :class:`SequenceFeature` for details on CPP feature concept.
        * :meth:`SequenceFeature.get_df_parts` for details on format of ``df_seq``.
        * The internally used :func:`seaborn.histplot` and :func:`seaborn.kdeplot` functions.

        Examples
        --------
        .. include:: examples/cpp_plot_feature.rst
        """
        # Check input
        ut.check_features(features=feature, list_scales=list(self._df_scales))
        ut.check_df_seq(df_seq=df_seq, accept_none=False)
        ut.check_number_val(name="label_test", val=label_test, just_int=True)
        ut.check_number_val(name="label_ref", val=label_ref, just_int=True)
        labels = ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref],
                                 len_requiered=len(df_seq), allow_other_vals=False)
        ut.check_ax(ax=ax, accept_none=True)
        ut.check_figsize(figsize=figsize, accept_none=True)
        names_to_show = ut.check_list_like(name="name_to_show", val=names_to_show, accept_str=True, accept_none=True)
        ut.check_str(name="name_test", val=name_test)
        ut.check_str(name="name_ref", val=name_ref)
        ut.check_color(name="color_test", val=color_test)
        ut.check_color(name="color_ref", val=color_ref)
        ut.check_bool(name="show_seq", val=show_seq)
        ut.check_bool(name="histplot", val=histplot)
        args_fs = ut.check_fontsize_args(fontsize_mean_dif=fontsize_mean_dif,
                                         fontsize_name_test=fontsize_name_test,
                                         fontsize_name_ref=fontsize_name_ref,
                                         fontsize_names_to_show=fontsize_names_to_show)
        ut.check_number_range(name="alpha_hist", val=alpha_hist, accept_none=False, min_val=0, max_val=1, just_int=False)
        ut.check_number_range(name="alpha_dif", val=alpha_dif, accept_none=False, min_val=0, max_val=1, just_int=False)
        check_match_df_seq_names_to_show(df_seq=df_seq, names_to_show=names_to_show)
        df_seq = check_match_df_seq_jmd_len(df_seq=df_seq, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len)

        # Plot feature
        ax = plot_feature(ax=ax, figsize=figsize,
                          feature=feature, df_scales=self._df_scales, accept_gaps=self._accept_gaps,
                          df_seq=df_seq, labels=labels, label_test=label_test, label_ref=label_ref,
                          jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                          names_to_show=names_to_show, show_seq=show_seq,
                          name_test=name_test, name_ref=name_ref,
                          color_test=color_test, color_ref=color_ref,
                          **args_fs,
                          histplot=histplot, alpha_hist=alpha_hist, alpha_dif=alpha_dif)
        return ax

    # Plotting methods for multiple features (group and sample level)
    def ranking(self,
                df_feat: pd.DataFrame = None,
                shap_plot: bool = False,
                col_dif: str = "mean_dif",
                col_imp: str = "feat_importance",
                rank: bool = True,
                n_top: int = 15,
                figsize: Tuple[Union[int, float], Union[int, float]] = (7, 5),
                tmd_len: int = 20,
                tmd_jmd_space: int = 2,
                tmd_color: str = "mediumspringgreen",
                jmd_color: str = "blue",
                tmd_jmd_alpha: Union[int, float] = 0.075,
                name_test: str = "TEST",
                name_ref: str = "REF",
                fontsize_titles: Union[int, float, None] = 12,
                fontsize_labels: Union[int, float, None] = 12,
                fontsize_annotations: Union[int, float, None] = 11,
                xlim_dif: Union[Tuple[Union[int, float], Union[int, float]], None] = (-17.5, 17.5),
                xlim_rank: Optional[Tuple[Union[int, float], Union[int, float]]] = (0, 4),
                rank_info_xy: Optional[Tuple[Optional[Union[int, float]], Optional[Union[int, float]]]] = None,
                ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot CPP/-SHAP feature ranking based on feature importance or sample-specif feature impact.

        Introduced in [Breimann24c]_, this method visualizes the most important features for discriminating between
        the test and the reference dataset groups. At sample level, the feature impact derived from SHAP values
        of a specific sample can be used for ranking if ``shap_plot=True`` and 'feature_impact' column in ``df_feat``.

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
            Must also include feature importance (`feat_importance`) or impact (``feat_impact_'name'``) columns.
        shap_plot : bool, default=False
            Set the analysis type: **CPP Analysis** (if ``False``) for group-level or
            **CPP-SHAP Analysis** for sample-level (or subgroup-level) results:

            **CPP Analysis**

            - ``col_dif``: Displays the group-level difference of feature values, with the `mean_dif` column selected by default.
            - ``col_imp``: Refers to the group-level `feat_importance` column (shown in gray) used for feature ranking.

            **CPP-SHAP Analysis**

            - ``col_dif``: Allows the selection of sample-specific differences against the reference group
              from a `mean_dif_'name'` column.
            - ``col_imp``: Enables the selection of specific feature impacts from a `feat_impact_'name'` column for
              an individual sample, where positive (red) and negative (blue) feature impacts are visualized in the ranking.

        col_dif : str, default='mean_dif'
            Column name in ``df_feat`` for differences in feature values. Must match with the ``shap_plot`` setting.
        col_imp : str, default='feat_importance'
            Column name in ``df_feat`` for feature importance/impact values. Must match with the ``shap_plot`` setting.
        rank : bool, default=True
            If ``True``, features will be ranked in descending order of ``col_imp`` values.
        n_top : int, default=15
            The number of top features to display. Should be 1 < ``n_top`` <= ``n_features``.
        figsize : tuple, default=(7, 5)
            Figure dimensions (width, height) in inches.
        tmd_len : int, default=20
            Length of TMD to be depicted (>0).
        tmd_jmd_space : int, default=2
            The space between TMD and JMD labels (>0) in the feature position subplot.
        tmd_color : str, default='mediumspringgreen'
            Color for TMD.
        jmd_color : str, default='blue'
            Color for JMDs.
        tmd_jmd_alpha : int or float, default=0.075
            The transparency alpha value [0-1] of the TMD-JMD area in the feature position subplot.
        name_test : str, default="TEST"
            Name of the test dataset to show in the mean difference subplot.
        name_ref : str, default="REF"
            Name of reference dataset to show in the mean difference subplot.
        fontsize_titles : int or float, default=12
            Font size of the titles.
        fontsize_labels : int or float , default=12
            Font size of plot labels.
        fontsize_annotations : int or float, default=11
            Font size of annotations.
        xlim_dif : tuple, default=(-17.5, 17.5)
            x-axis limits for the mean difference subplot.
        xlim_rank : tuple, default=(0, 4)
            x-axis limits for the ranking subplot. If ``None``, determined automatically.
        rank_info_xy : tuple, optional
            Position (x-axis, y-axis) in ranking subplot for showing additional information (optimized if ``None``):

            - When ``shap_plot=False``: Displays sum of feature importance.
            - When ``shap_plot=True``: Show the sum of the absolute feature impact and the SHAP legend.

        Returns
        -------
        fig : plt.Figure
            The Figure object for the ranking plot.
        axes : array of plt.Axes
            Array of Axes objects, each representing a subplot within the figure.

        Notes
        -----
        * Features are shown as ordered in ``df_feat``. A ranking in descending order based one the following
          columns is recommended:

          - ``feat_importance``: when feature importance is in ``df_feat`` and ``shap_plot=False``.
          - ``feat_impact_'name'``: when sample-specific feature impact is in ``df_feat`` and ``shap_plot=True``.

        See Also
        --------
        * :meth:`CPP.run` for details on CPP statistical measures of the ``df_feat`` DataFrame.
        * :class:`SequenceFeature` for definition of sequence ``Parts``.
        * :meth:`CPPPlot.feature` for visualization of mean differences for specific features.

        Examples
        --------
        .. include:: examples/cpp_plot_ranking.rst
        """
        # Check input
        ut.check_bool(name="shap_plot", val=shap_plot)
        check_col_dif(col_dif=col_dif, shap_plot=shap_plot)
        col_imp = check_col_imp(col_imp=col_imp, shap_plot=shap_plot)
        df_feat = ut.check_df_feat(df_feat=df_feat, shap_plot=shap_plot, cols_requiered=[col_imp, col_dif])
        ut.check_number_range(name="n_top", val=n_top, min_val=2, max_val=len(df_feat), just_int=True)
        ut.check_bool(name="rank", val=rank, accept_none=False)
        ut.check_figsize(figsize=figsize, accept_none=True)
        args_len, _ = check_parts_len(tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len)
        check_match_features_seq_parts(features=df_feat[ut.COL_FEATURE],
                                       tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len)
        ut.check_number_range(name="tmd_jmd_space", val=tmd_jmd_space, min_val=1, just_int=True, accept_none=False)
        ut.check_color(name="tmd_color", val=tmd_color)
        ut.check_color(name="jmd_color", val=jmd_color)
        ut.check_number_range(name="tmd_jmd_alpha", val=tmd_jmd_alpha, accept_none=False, min_val=0, max_val=1, just_int=False)
        ut.check_str(name="name_test", val=name_test)
        ut.check_str(name="name_ref", val=name_ref)
        args = dict(min_val=0, exclusive_limits=True, accept_none=True, just_int=False)
        ut.check_number_range(name="fontsize_titles", val=fontsize_titles, **args)
        ut.check_number_range(name="fontsize_labels", val=fontsize_labels, **args)
        ut.check_number_range(name="fontsize_annotations", val=fontsize_annotations, **args)
        ut.check_lim(name="xlim_dif", val=xlim_dif)
        ut.check_lim(name="xlim_rank", val=xlim_rank)
        ut.check_tuple(name="rank_info_xy", val=rank_info_xy, n=2,
                       accept_none=True, check_number=True)

        # DEV: No match check for features and tmd (check_match_features_seq_parts) necessary
        # Plot ranking
        fig, axes = plot_ranking(df_feat=df_feat.copy(),
                                 n_top=n_top, rank=rank,
                                 col_dif=col_dif,
                                 col_imp=col_imp,
                                 shap_plot=shap_plot,
                                 figsize=figsize,
                                 tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                 tmd_color=tmd_color, jmd_color=jmd_color,
                                 tmd_jmd_alpha=tmd_jmd_alpha,
                                 name_test=name_test, name_ref=name_ref,
                                 fontsize_titles=fontsize_titles,
                                 fontsize_labels=fontsize_labels,
                                 fontsize_annotations=fontsize_annotations,
                                 tmd_jmd_space=tmd_jmd_space,
                                 xlim_dif=xlim_dif, xlim_rank=xlim_rank,
                                 rank_info_xy=rank_info_xy)

        # Adjust plot
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            fig.tight_layout()
            plt.subplots_adjust(left=0.25, wspace=0.15)
        return fig, axes


    def profile(self,
                # Data and Plot Type
                df_feat: pd.DataFrame = None,
                shap_plot: bool = False,
                col_imp: Union[str, None] = "feat_importance",
                normalize: bool = True,
                ax: Optional[plt.Axes] = None,
                figsize: Tuple[Union[int, float], Union[int, float]] = (7, 5),

                # Appearance of Parts (TMD-JMD)
                start: int = 1,
                tmd_len: int = 20,
                tmd_seq: Optional[str] = None,
                jmd_n_seq: Optional[str] = None,
                jmd_c_seq: Optional[str] = None,
                tmd_color: str = "mediumspringgreen",
                jmd_color: str = "blue",
                tmd_seq_color: str = "black",
                jmd_seq_color: str = "white",
                seq_size: Union[int, float] = None,
                fontsize_tmd_jmd: Union[int, float] = None,
                weight_tmd_jmd: Literal['normal', 'bold'] = "normal",
                add_xticks_pos: bool = False,
                highlight_tmd_area: bool = True,
                highlight_alpha: float = 0.15,

                # Legend, Axis, and Grid Configurations
                add_legend_cat: bool = False,
                dict_color: Optional[dict] = None,
                legend_kws: Optional[dict] = None,
                bar_width: Union[int, float] = 0.75,
                edge_color: Optional[str] = None,
                grid_axis: Optional[Literal['x', 'y', 'both']] = None,
                ylim: Tuple[float, float] = None,
                xtick_size: Union[int, float] = 11.0,
                xtick_width: Union[int, float] = 2.0,
                xtick_length: Union[int, float] = 5.0,
                ytick_size: Optional[Union[int, float]] = None,
                ytick_width: Optional[Union[int, float]] = None,
                ytick_length: Union[int, float] = 5.0,
                ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot CPP/-SHAP profile showing feature importance/impact per residue position.

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
            Must also include either ``feat_importance`` or ``feat_impact`` column.
        shap_plot : bool, default=False
            Set the analysis type: **CPP Analysis** (if ``False``) for group-level or
            **CPP-SHAP Analysis** for sample-level (or subgroup-level) results:

             **CPP Analysis**

             - ``col_imp``: Refers to the group-level `feat_importance` column (shown in gray),
               depicted by gray bars for each residue position.

             **CPP-SHAP Analysis**

             - ``col_imp``: Enables the selection of specific feature impacts from a `feat_impact_'name'` column for
               an individual sample, where positive (red) and negative (blue) feature impacts are visualized by +/- bars.

        col_imp : str or None, default='feat_importance'
            Column name in ``df_feat`` for feature importance/impact values. Must match with the ``shap_plot`` setting.
            If ``None``, the number of features per residue position will be shown.
        normalize : bool, default=True
            If ``True``, normalizes aggregated numerical values to a total of 100%.
        ax : plt.Axes, optional
            Pre-defined Axes object to plot on. If ``None``, a new Axes object is created.
        figsize : tuple, default=(7, 5)
            Figure dimensions (width, height) in inches.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD to be depicted (>0). Must match with all feature from ``df_feat``.
        tmd_seq : str, optional
            TMD sequence for specific sample.
        jmd_n_seq : str, optional
            JMD N-terminal sequence for specific sample. Length must match with 'jmd_n_len' attribute.
        jmd_c_seq : str, optional
            JMD C-terminal sequence for specific sample. Length must match with 'jmd_c_len' attribute.
        tmd_color : str, default='mediumspringgreen'
            Color for TMD.
        jmd_color : str, default='blue'
            Color for JMDs.
        tmd_seq_color : str, default='black'
            Color for TMD sequence.
        jmd_seq_color : str, default='white'
            Color for JMD sequence.
        seq_size : int or float, optional
            Font size (>=0) for sequence characters. If ``None``, optimized automatically.
        fontsize_tmd_jmd : int or float, optional
            Font size (>=0) for the part labels: 'JMD-N', 'TMD', 'JMD-C'. If ``None``, optimized automatically.
        weight_tmd_jmd : {'normal', 'bold'}, default='normal'
            Font weight for the part labels: 'JMD-N', 'TMD', 'JMD-C'.
        add_xticks_pos : bool, default=False
            If ``True``, include x-tick positions when TMD-JMD sequence is given.
        highlight_tmd_area : bool, default=True
            If ``True``, highlights the TMD area on the plot.
        highlight_alpha : float, default=0.15
            The transparency alpha value [0-1] for TMD area highlighting.
        add_legend_cat : bool, default=False
            If ``True``, the scale categories are indicated as stacked bars and a legend is added. If ``True``,
            ensure that ``shap_plot=False``.
        dict_color : dict, optional
            Color dictionary of scale categories for legend. Default from :meth:`plot_get_cdict` with ``name='DICT_CAT'``.
        legend_kws : dict, optional
            Keyword arguments for the legend passed to :meth:`plot_legend`.
        bar_width : int or float, default=0.75
            Width of the bars.
        edge_color : str, optional
            Color of the bar edges.
        grid_axis : {'x', 'y', 'both', None}, default=None
            Axis on which the grid is drawn if not ``None``.
        ylim : tuple, optional
            Y-axis limits. If ``None``, y-axis limits are set automatically.
        xtick_size : int or float, default=11.0
            Size of x-tick labels (>0).
        xtick_width : int or float, default=2.0
            Width of the x-ticks (>0).
        xtick_length : int or float, default=5.0
            Length of the x-ticks (>0).
        ytick_size : int or float, optional
            Size of y-tick labels (>0).
        ytick_width : int or float, default=2.0
            Width of the y-ticks (>0).
        ytick_length : int or float, default=5.0
            Length of the y-ticks (>0).

        Returns
        -------
        fig : plt.Figure
            The Figure object for the CPP profile plot.
        ax : plt.Axes
            CPP profile plot axes object.

        Notes
        -----
        * ``tmd_seq_color`` and ``jmd_seq_color`` are applicable only when ``tmd_seq``, ``jmd_n_seq``,
          and ``jmd_c_seq`` are provided.

        Warnings
        --------
        * If ``ylim`` does not match with minimum and/or maximum of aggregate numerical values across all residue
          position, a ``UserWarning`` is raised and ``ylim`` will be adjusted automatically.

        Examples
        --------
        .. include:: examples/cpp_plot_profile.rst
        """
        # Check primary input
        ut.check_bool(name="shap_plot", val=shap_plot)
        if col_imp is None:
            df_feat = ut.check_df_feat(df_feat=df_feat, df_cat=self._df_cat)
        else:
            col_imp = check_col_imp(col_imp=col_imp, shap_plot=shap_plot)
            df_feat = ut.check_df_feat(df_feat=df_feat, df_cat=self._df_cat, shap_plot=shap_plot, cols_requiered=[col_imp])
        ut.check_bool(name="normalize", val=normalize)
        ut.check_ax(ax=ax, accept_none=True)
        ut.check_figsize(figsize=figsize, accept_none=True)

        # Check specific TMD-JMD input
        ut.check_number_range(name="start", val=start, min_val=0, just_int=True)
        args_len, args_seq = check_parts_len(tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                             jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq,
                                             check_jmd_seq_len_consistent=True)
        args_part_color = check_part_color(tmd_color=tmd_color, jmd_color=jmd_color)
        args_seq_color = check_seq_color(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
        args_fs = ut.check_fontsize_args(seq_size=seq_size,
                                         fontsize_tmd_jmd=fontsize_tmd_jmd)
        ut.check_font_weight(name="weight_tmd_jmd", font_weight=weight_tmd_jmd)
        ut.check_bool(name="add_xticks_pos", val=add_xticks_pos)
        ut.check_bool(name="highlight_tmd_area", val=highlight_tmd_area)
        ut.check_number_range(name="tmd_area_alpha", val=highlight_alpha, min_val=0, max_val=1, just_int=False)
        check_match_features_seq_parts(features=df_feat[ut.COL_FEATURE],
                                       tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                       tmd_seq=tmd_seq, jmd_n_seq=jmd_n_seq, jmd_c_seq=jmd_c_seq)
        check_match_shap_plot_add_legend_cat(shap_plot=shap_plot, add_legend_cat=add_legend_cat)

        # Check plot styling input
        ut.check_bool(name="add_legend_cat", val=add_legend_cat)
        ut.check_dict(name="legend_kws", val=legend_kws, accept_none=True)
        dict_color = check_match_dict_color_df(dict_color=dict_color, df=df_feat)
        ut.check_number_range(name="bar_width", val=bar_width, min_val=0, just_int=False)
        ut.check_color(name="edge_color", val=edge_color, accept_none=True)
        ut.check_grid_axis(grid_axis=grid_axis)
        ut.check_lim(name="ylim", val=ylim, accept_none=True)
        args_xtick = check_args_xtick(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)
        args_ytick = check_args_ytick(ytick_size=ytick_size, ytick_width=ytick_width, ytick_length=ytick_length)

        # Plot profile
        fig, ax = plot_profile(df_feat=df_feat, df_cat=self._df_cat, shap_plot=shap_plot,
                               col_imp=col_imp, normalize=normalize,
                               figsize=figsize, ax=ax,
                               start=start, **args_len, **args_seq,
                               **args_part_color, **args_seq_color,
                               **args_fs, weight_tmd_jmd=weight_tmd_jmd,
                               add_xticks_pos=add_xticks_pos,
                               highlight_tmd_area=highlight_tmd_area, highlight_alpha=highlight_alpha,
                               add_legend_cat=add_legend_cat, dict_color=dict_color, legend_kws=legend_kws,
                               bar_width=bar_width, edge_color=edge_color,
                               grid_axis=grid_axis, ylim=ylim, **args_xtick, **args_ytick)

        # Adjust plot
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            plt.tight_layout()
        if tmd_seq is not None and seq_size is None:
            ax, seq_size = update_seq_size_(ax=ax, **args_seq, **args_part_color, **args_seq_color)
            if self._verbose:
                ut.print_out(f"Optimized sequence character fontsize is: {seq_size}")
        return fig, ax


    def heatmap(self,
                # Data and Plot Type
                df_feat: pd.DataFrame = None,
                shap_plot: bool = False,
                col_cat : Literal['category', 'subcategory', 'scale_name'] = "subcategory",
                col_val: str = "mean_dif",
                name_test: str = "TEST",
                name_ref: str = "REF",
                figsize: Tuple[Union[int, float], Union[int, float]] = (8, 8),

                # Appearance of Parts (TMD-JMD)
                start: int = 1,
                tmd_len: int = 20,
                tmd_seq: Optional[str] = None,
                jmd_n_seq: Optional[str] = None,
                jmd_c_seq: Optional[str] = None,
                tmd_color: str = "mediumspringgreen",
                jmd_color: str = "blue",
                tmd_seq_color: str = "black",
                jmd_seq_color: str = "white",
                seq_size: Optional[Union[int, float]] = None,
                fontsize_tmd_jmd: Optional[Union[int, float]] = None,
                weight_tmd_jmd: Literal['normal', 'bold'] = "normal",
                fontsize_labels: Union[int, float] = 12,
                add_xticks_pos: bool = False,

                # Legend, Axis, and Grid Configurations
                grid_linewidth: Union[int, float] = 0.01,
                grid_linecolor: Optional[str] = None,
                border_linewidth: Union[int, float] = 2,
                facecolor_dark: Optional[bool] = None,
                vmin: Optional[Union[int, float]] = None,
                vmax: Optional[Union[int, float]] = None,
                cmap: Optional[str] = None,
                cmap_n_colors: int = 101,
                cbar_pct: bool = True,
                cbar_kws: Optional[dict] = None,
                cbar_xywh: Tuple[Optional[float], Optional[float], Optional[float], Optional[float]] = (0.7, None, 0.2, None),
                dict_color: Optional[dict] = None,
                legend_kws: Optional[dict] = None,
                legend_xy: Tuple[Optional[float], Optional[float]] = (-0.1, -0.01),
                xtick_size: Union[int, float] = 11.0,
                xtick_width: Union[int, float] = 2.0,
                xtick_length: Union[int, float] = 5.0,
                ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot a CPP/-SHAP heatmap showing the feature value mean difference/feature impact
        per scale subcategory (y-axis) and residue position (x-axis).

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_feature, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
            Can also include feature impact (``feat_impact``) column.
        shap_plot : bool, default=False
            Set the analysis type: **CPP Analysis** (if ``False``) for group-level or
            **CPP-SHAP Analysis** for sample-level (or subgroup-level) results:

             **CPP Analysis**

            - ``col_val``: Displays typically the difference of feature values, either at group-level when the `mean_dif`
              column is selected or at sample-level (group-level) when a `mean_dif_'name'` column is provided.

            **CPP-SHAP Analysis**

            - ``col_val``: Enables typically the selection of specific feature impacts from a `feat_impact_'name'` column
              for an individual sample, where positive (red) and negative (blue) feature impacts are indicated.

        col_cat : {'category', 'subcategory', 'scale_name'}, default='subcategory'
            Column name in ``df_feat`` for scale classification (y-axis).
        col_val : {'mean_dif', 'abs_mean_dif', 'abs_auc', 'feat_importance', ``mean_dif_'name'``, ``feat_impact_'name'``}, default='mean_dif'
            Column name in ``df_feat`` for numerical values to display. Must match with the ``shap_plot`` setting.
        name_test : str, default="TEST"
            Name for the test dataset.
        name_ref : str, default="REF"
            Name for the reference dataset.
        figsize : tuple, default=(8, 8)
            Figure dimensions (width, height) in inches.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD to be depicted (>0). Must match with all feature from ``df_feat``.
        tmd_seq : str, optional
            TMD sequence for specific sample.
        jmd_n_seq : str, optional
            JMD N-terminal sequence for specific sample. Length must match with 'jmd_n_len' attribute.
        jmd_c_seq : str, optional
            JMD C-terminal sequence for specific sample. Length must match with 'jmd_c_len' attribute.
        tmd_color : str, default='mediumspringgreen'
            Color for TMD.
        jmd_color : str, default='blue'
            Color for JMDs.
        tmd_seq_color : str, default='black'
            Color for TMD sequence.
        jmd_seq_color : str, default='white'
            Color for JMD sequence.
        seq_size : int or float, optional
            Font size (>=0) for sequence characters. If ``None``, optimized automatically.
        fontsize_tmd_jmd : int or float, optional
            Font size (>=0) for the part labels: 'JMD-N', 'TMD', 'JMD-C'. If ``None``, optimized automatically.
        weight_tmd_jmd : {'normal', 'bold'}, default='normal'
            Font weight for the part labels: 'JMD-N', 'TMD', 'JMD-C'.
        fontsize_labels : int or float, default=12
            Font size (>= 0) for figure labels. If ``None``, determined automatically.
        add_xticks_pos : bool, default=False
            If ``True``, include x-tick positions when TMD-JMD sequence is given.
        grid_linewidth : int or float, default=0.01
            Line width for the grid.
        grid_linecolor : str, optional
            Color for the grid lines. If ``None``, automatically determined based on ``facecolor_dark``.
        border_linewidth : int or float, default=2
            Line width for the TMD-JMD border.
        facecolor_dark : bool, optional
            Sets background of heatmap to black (if ``True``) or white. If ``None``, automatically determined from
            ``shap_plot`` setting. Affects grid cells for missing or near-zero data based on ``col_val``.
        vmin : int or float, optional
            Minimum ``col_val`` value setting the lower end of the colormap. If ``None``, determined automatically.
        vmax : int or float, optional
            Maximum ``col_val`` value setting the upper end of the colormap. If ``None``, determined automatically.
        cmap : matplotlib colormap name or object, optional
            Name of the colormap to use. If ``None``, automatically determined ``col_val`` data and 'shap_plot' setting.
        cmap_n_colors : int, default=101
            Number of discrete steps (>1) in diverging or sequential colormap.
        cbar_pct : bool, default=True
            If ``True``, colorbar is represented in percentage and the ``col_val`` values are converted
            accordingly by multiplying with 100 if necessary.
        cbar_kws : dict of key, value mappings, optional
            Keyword arguments for colorbar passed to :meth:`matplotlib.figure.Figure.colorbar`.
        cbar_xywh : tuple, default=(0.7, None, 0.2, None)
            Colorbar position and size: x-axis (left), y-axis (bottom), width, height. Values are optimized if ``None``.
        dict_color : dict, optional
            Color dictionary of scale categories classifying scales shown on y-axis. Default from
            :meth:`plot_get_cdict` with ``name='DICT_CAT'``.
        legend_kws : dict, optional
            Keyword arguments for the legend passed to :meth:`plot_legend`.
        legend_xy : tuple, default=(-0.1, -0.01)
            Position for scale category legend: x- and y-axis coordinates. Values are set to default if ``None``.
        xtick_size : int or float, default=11.0
            Size of x-tick labels (>0).
        xtick_width : int or float, default=2.0
            Width of the x-ticks (>0).
        xtick_length : int or float, default=5.0
            Length of the x-ticks (>0).

        Returns
        -------
        fig : plt.Figure
            The Figure object for the CPP heatmap.
        ax : plt.Axes
            Array of Axes objects for the CPP heatmap.

        Notes
        -----
        ``tmd_seq_color`` and ``jmd_seq_color`` are applicable only when ``tmd_seq``, ``jmd_n_seq``,
        and ``jmd_c_seq`` are provided.

        See Also
        --------
        * :meth:`CPP.run` for details on CPP statistical measures of the ``df_feat`` DataFrame.
        * :class:`SequenceFeature` for definition of sequence ``Parts``.
        * :meth:`CPPPlot.feature` for visualization of mean differences for specific features.
        * :meth:`seaborn.heatmap` for seaborn heatmap.
        * :meth:`matplotlib.figure.Figure.colorbar` for colorbar arguments.
        * `Matplotlib Colormaps <https://matplotlib.org/stable/users/explain/colors/colormaps.html>`_ for further ``cmap`` options.
        * :meth:`plot_legend` used for setting scale category legend.

        Examples
        --------
        .. include:: examples/cpp_plot_heatmap.rst
        """
        # Check primary input
        ut.check_bool(name="shap_plot", val=shap_plot)
        check_col_cat(col_cat=col_cat)
        col_val = check_col_val(col_val=col_val, shap_plot=shap_plot)
        df_feat = ut.check_df_feat(df_feat=df_feat, df_cat=self._df_cat, shap_plot=shap_plot,
                                   cols_requiered=col_val, cols_nan_check=col_val)
        ut.check_str(name="name_test", val=name_test)
        ut.check_str(name="name_ref", val=name_ref)
        ut.check_figsize(figsize=figsize, accept_none=True)

        # Check specific TMD-JMD input
        ut.check_number_range(name="start", val=start, min_val=0, just_int=True)
        args_len, args_seq = check_parts_len(tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                             jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq,
                                             check_jmd_seq_len_consistent=True)
        args_part_color = check_part_color(tmd_color=tmd_color, jmd_color=jmd_color)
        args_seq_color = check_seq_color(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
        args_fs = ut.check_fontsize_args(seq_size=seq_size,
                                         fontsize_tmd_jmd=fontsize_tmd_jmd,
                                         fontsize_labels=fontsize_labels)
        ut.check_font_weight(name="weight_tmd_jmd", font_weight=weight_tmd_jmd)
        ut.check_bool(name="add_xticks_pos", val=add_xticks_pos)
        check_match_features_seq_parts(features=df_feat[ut.COL_FEATURE],
                                       tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                       tmd_seq=tmd_seq, jmd_n_seq=jmd_n_seq, jmd_c_seq=jmd_c_seq)

        # Check plot styling input
        ut.check_number_range(name="grid_linewidth", val=grid_linewidth, min_val=0, just_int=False)
        ut.check_color(name="grid_linecolor", val=grid_linecolor, accept_none=True)
        ut.check_number_range(name="border_linewidth", val=border_linewidth, min_val=0, just_int=False)
        ut.check_bool(name="facecolor_dark", val=facecolor_dark, accept_none=True)
        ut.check_vmin_vmax(vmin=vmin, vmax=vmax)
        check_cmap_for_heatmap(cmap=cmap)
        ut.check_number_range(name="cmap_n_colors", val=cmap_n_colors, min_val=2, accept_none=True, just_int=True)
        ut.check_bool(name="cbar_pct", val=cbar_pct)
        ut.check_dict(name="cbar_kws", val=cbar_kws, accept_none=True)
        ut.check_tuple(name="cbar_xywh", val=cbar_xywh, n=4, accept_none=False,
                       check_number=True, accept_none_number=True)
        dict_color = check_match_dict_color_df(dict_color=dict_color, df=df_feat)
        ut.check_dict(name="legend_kws", val=legend_kws, accept_none=True)
        ut.check_tuple(name="legend_xy", val=legend_xy, n=2, accept_none=False,
                       check_number=True, accept_none_number=True)
        args_xtick = check_args_xtick(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)

        # Plot heatmap
        fig, ax = plot_heatmap(df_feat=df_feat, df_cat=self._df_cat,
                               shap_plot=shap_plot,
                               col_cat=col_cat, col_val=col_val,
                               name_test=name_test, name_ref=name_ref,
                               figsize=figsize,
                               start=start, **args_len, **args_seq,
                               **args_part_color, **args_seq_color,
                               **args_fs, weight_tmd_jmd=weight_tmd_jmd,
                               add_xticks_pos=add_xticks_pos,
                               grid_linewidth=grid_linewidth, grid_linecolor=grid_linecolor,
                               border_linewidth=border_linewidth,
                               facecolor_dark=facecolor_dark, vmin=vmin, vmax=vmax,
                               cmap=cmap, cmap_n_colors=cmap_n_colors,
                               cbar_pct=cbar_pct, cbar_kws=cbar_kws, cbar_xywh=cbar_xywh,
                               dict_color=dict_color, legend_kws=legend_kws, legend_xy=legend_xy,
                               **args_xtick)

        # Adjust plot
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            fig.tight_layout()
        if tmd_seq is not None and seq_size is None:
            ax, seq_size = update_seq_size_(ax=ax, **args_seq, **args_part_color, **args_seq_color)
            if self._verbose:
                ut.print_out(f"Optimized sequence character fontsize is: {seq_size}")
        return fig, ax


    def feature_map(self,
                    # Data and Plot Type
                    df_feat: pd.DataFrame = None,
                    col_cat: Literal['category', 'subcategory', 'scale_name'] = "subcategory",
                    col_val: str = "mean_dif",
                    col_imp: str = "feat_importance",
                    name_test: str = "TEST",
                    name_ref: str = "REF",
                    figsize: Tuple[Union[int, float], Union[int, float]] = (8, 8),

                    # Appearance of Parts (TMD-JMD)
                    start: int = 1,
                    tmd_len: int = 20,
                    tmd_seq: Optional[str] = None,
                    jmd_n_seq: Optional[str] = None,
                    jmd_c_seq: Optional[str] = None,
                    tmd_color: str = "mediumspringgreen",
                    jmd_color: str = "blue",
                    tmd_seq_color: str = "black",
                    jmd_seq_color: str = "white",
                    seq_size: Optional[Union[int, float]] = None,
                    fontsize_tmd_jmd: Optional[Union[int, float]] = None,
                    weight_tmd_jmd: Literal['normal', 'bold'] = "normal",
                    fontsize_titles: Union[int, float] = 11,
                    fontsize_labels: Union[int, float] = 12,
                    fontsize_annotations: Union[int, float] = 11,
                    fontsize_imp_bar: Union[int, float] = 9,
                    add_xticks_pos: bool = False,

                    # Legend, Axis, and Grid Configurations
                    grid_linewidth: Union[int, float] = 0.01,
                    grid_linecolor: Optional[str] = None,
                    border_linewidth: Union[int, float] = 2,
                    facecolor_dark: bool = False,
                    vmin: Optional[Union[int, float]] = None,
                    vmax: Optional[Union[int, float]] = None,
                    cmap: Optional[str] = None,
                    cmap_n_colors: int = 101,
                    cbar_pct: bool = True,
                    cbar_kws: Optional[dict] = None,
                    cbar_xywh: Tuple[Optional[float], Optional[float], Optional[float], Optional[float]] = (0.5, None, 0.2, None),
                    dict_color: Optional[dict] = None,
                    legend_kws: Optional[dict] = None,
                    legend_xy: Tuple[Optional[float], Optional[float]] = (-0.1, -0.01),
                    legend_imp_xy: Tuple[Optional[float], Optional[float]] = (1.25, 0),
                    imp_ths: Tuple[Optional[float], Optional[float], Optional[float]] = (0.2, 0.5, 1),
                    imp_marker_sizes: Tuple[Optional[float], Optional[float], Optional[float]] = (3, 5, 8),
                    imp_bar_th: Optional[Union[int, float]] = None,
                    xtick_size: Union[int, float] = 11.0,
                    xtick_width: Union[int, float] = 2.0,
                    xtick_length: Union[int, float] = 5.0,
                    )-> Tuple[plt.Figure, plt.Axes]:
        """
        Plot CPP feature map showing feature value mean difference and feature importance
        per scale subcategory (y-axis) and residue position (x-axis).

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_feature, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
            Must also include a feature importance column (``col_imp``).
        col_cat : {'category', 'subcategory', 'scale_name'}, default='subcategory'
            Column name in ``df_feat`` for scale information (y-axis).
        col_val : {'mean_dif', 'abs_mean_dif', 'abs_auc'}, default='mean_dif'
            Column name in ``df_feat`` for numerical values to display.
        col_imp :  {``feat_importance``, ``feat_importance_'name'``}, default='feat_importance'
            Column name in ``df_feat`` for feature importance (group-, subgroup- or sample-level).
        name_test : str, default="TEST"
            Name for the test dataset.
        name_ref : str, default="REF"
            Name for the reference dataset.
        figsize : tuple, default=(8, 8)
            Figure dimensions (width, height) in inches.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD to be depicted (>0). Must match with all feature from ``df_feat``.
        tmd_seq : str, optional
            TMD sequence for specific sample.
        jmd_n_seq : str, optional
            JMD N-terminal sequence for specific sample. Length must match with 'jmd_n_len' attribute.
        jmd_c_seq : str, optional
            JMD C-terminal sequence for specific sample. Length must match with 'jmd_c_len' attribute.
        tmd_color : str, default='mediumspringgreen'
            Color for TMD.
        jmd_color : str, default='blue'
            Color for JMDs.
        tmd_seq_color : str, default='black'
            Color for TMD sequence.
        jmd_seq_color : str, default='white'
            Color for JMD sequence.
        seq_size : int or float, optional
            Font size (>=0) for sequence characters. If ``None``, optimized automatically.
        fontsize_tmd_jmd : int or float, optional
            Font size (>=0) for the part labels: 'JMD-N', 'TMD', 'JMD-C'. If ``None``, optimized automatically.
        weight_tmd_jmd : {'normal', 'bold'}, default='normal'
            Font weight for the part labels: 'JMD-N', 'TMD', 'JMD-C'.
        fontsize_titles : int or float, default=11
            Font size (>= 0) for figure titles. If ``None``, determined automatically.
        fontsize_labels : int or float, default=12
            Font size (>= 0) for figure labels. If ``None``, determined automatically.
        fontsize_annotations : int or float, default=10
            Font size (>= 0) for figure annotations. If ``None``, determined automatically.
        fontsize_imp_bar : int or float, default=9
            Font size (>= 0) for feature importance in bars. If ``None``, determined automatically.
        add_xticks_pos : bool, default=False
            If ``True``, include x-tick positions when TMD-JMD sequence is given.
        grid_linewidth : int or float, default=0.01
            Line width for the grid.
        grid_linecolor : str, optional
            Color for the grid lines. If ``None``, automatically determined based on ``facecolor_dark``.
        border_linewidth : int or float, default=2
            Line width for the TMD-JMD border.
        facecolor_dark : bool, optional
            Sets background of heatmap to black (if ``True``) or white. If ``None``, automatically determined from
            ``shap_plot`` setting. Affects grid cells for missing or near-zero data based on ``col_val``.
        vmin : int or float, optional
            Minimum ``col_val`` value setting the lower end of the colormap. If ``None``, determined automatically.
        vmax : int or float, optional
            Maximum ``col_val`` value setting the upper end of the colormap. If ``None``, determined automatically.
        cmap : matplotlib colormap name or object, optional
            Name of the colormap to use. If ``None``, automatically determined ``col_val`` data and 'shap_plot' setting.
        cmap_n_colors : int, default=101
            Number of discrete steps (>1) in diverging or sequential colormap.
        cbar_pct : bool, default=True
            If ``True``, colorbar is represented in percentage and the ``col_val`` values are converted
            accordingly by multiplying with 100 if necessary.
        cbar_kws : dict of key, value mappings, optional
            Keyword arguments for colorbar passed to :meth:`matplotlib.figure.Figure.colorbar`.
        cbar_xywh : tuple, default=(0.7, None, 0.2, None)
            Colorbar position and size: x-axis (left), y-axis (bottom), width, height. Values are optimized if ``None``.
        dict_color : dict, optional
            Color dictionary of scale categories classifying scales shown on y-axis. Default from
            :meth:`plot_get_cdict` with ``name='DICT_CAT'``.
        legend_kws : dict, optional
            Keyword arguments for the legend passed to :meth:`plot_legend`.
        legend_xy : tuple, default=(-0.1, -0.01)
            Position for scale category legend: x- and y-axis coordinates. Values are set to default if ``None``.
        legend_imp_xy : tuple, default=(1.25, 0)
            Position for feature importance legend: x- and y-axis coordinates (relative to cbar).
        imp_ths : tuple, default=(0.2, 0.5, 1)
            Three ascending thresholds for feature importance (scale- and residue-specific).
        imp_marker_sizes : tuple, default=(3, 5, 8)
            Size of three feature importance markers defined by ``impd_th``.
        imp_bar_th : int or float, optional
            Threshold for cumulated feature importance to be shown on right bars. If ``None``, determined automatically.
        xtick_size : int or float, default=11.0
            Size of x-tick labels (>0).
        xtick_width : int or float, default=2.0
            Width of the x-ticks (>0).
        xtick_length : int or float, default=5.0
            Length of the x-ticks (>0).

        Returns
        -------
        fig : plt.Figure
            The Figure object for the CPP feature map.
        ax : plt.Axes
            Array of Axes objects for the CPP feature map.

        Notes
        -----
        ``tmd_seq_color`` and ``jmd_seq_color`` are applicable only when ``tmd_seq``, ``jmd_n_seq``,
        and ``jmd_c_seq`` are provided.

        See Also
        --------
        * :meth:`CPP.run` for details on CPP statistical measures of the ``df_feat`` DataFrame.
        * :class:`SequenceFeature` for definition of sequence ``Parts``.
        * :meth:`CPPPlot.feature` for visualization of mean differences for specific features.
        * :meth:`seaborn.heatmap` for seaborn heatmap.
        * :meth:`matplotlib.figure.Figure.colorbar` for colorbar arguments.
        * `Matplotlib Colormaps <https://matplotlib.org/stable/users/explain/colors/colormaps.html>`_ for further ``cmap`` options.
        * :meth:`plot_legend` used for setting scale category legend.

        Examples
        --------
        .. include:: examples/cpp_plot_feature_map.rst
        """
        # Check primary input
        check_col_cat(col_cat=col_cat)
        col_val = check_col_val(col_val=col_val, sample_mean_dif=True)
        col_imp = check_col_imp(col_imp=col_imp)
        df_feat = ut.check_df_feat(df_feat=df_feat, df_cat=self._df_cat,
                                   cols_requiered=[col_val, col_imp],
                                   cols_nan_check=col_val)

        ut.check_str(name="name_test", val=name_test)
        ut.check_str(name="name_ref", val=name_ref)
        ut.check_figsize(figsize=figsize, accept_none=True)

        # Check specific TMD-JMD input
        ut.check_number_range(name="start", val=start, min_val=0, just_int=True)
        args_len, args_seq = check_parts_len(tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                             jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq,
                                             check_jmd_seq_len_consistent=True)
        args_part_color = check_part_color(tmd_color=tmd_color, jmd_color=jmd_color)
        args_seq_color = check_seq_color(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
        args_fs = ut.check_fontsize_args(seq_size=seq_size,
                                         fontsize_tmd_jmd=fontsize_tmd_jmd,
                                         fontsize_titles=fontsize_titles,
                                         fontsize_labels=fontsize_labels,
                                         fontsize_annotations=fontsize_annotations,
                                         fontsize_imp_bar=fontsize_imp_bar)
        ut.check_font_weight(name="weight_tmd_jmd", font_weight=weight_tmd_jmd)
        ut.check_bool(name="add_xticks_pos", val=add_xticks_pos)
        check_match_features_seq_parts(features=df_feat[ut.COL_FEATURE],
                                       tmd_len=tmd_len, jmd_n_len=self._jmd_n_len, jmd_c_len=self._jmd_c_len,
                                       tmd_seq=tmd_seq, jmd_n_seq=jmd_n_seq, jmd_c_seq=jmd_c_seq)

        # Check plot styling input
        ut.check_number_range(name="grid_linewidth", val=grid_linewidth, min_val=0, just_int=False)
        ut.check_color(name="grid_linecolor", val=grid_linecolor, accept_none=True)
        ut.check_number_range(name="border_linewidth", val=border_linewidth, min_val=0, just_int=False)
        ut.check_bool(name="facecolor_dark", val=facecolor_dark, accept_none=False)
        ut.check_vmin_vmax(vmin=vmin, vmax=vmax)
        check_cmap_for_heatmap(cmap=cmap)
        ut.check_number_range(name="cmap_n_colors", val=cmap_n_colors, min_val=2, accept_none=True, just_int=True)
        ut.check_bool(name="cbar_pct", val=cbar_pct)
        ut.check_dict(name="cbar_kws", val=cbar_kws, accept_none=True)
        ut.check_tuple(name="cbar_xywh", val=cbar_xywh, n=4, accept_none=False,
                       check_number=True, accept_none_number=True)
        dict_color = check_match_dict_color_df(dict_color=dict_color, df=df_feat)
        ut.check_dict(name="legend_kws", val=legend_kws, accept_none=True)
        ut.check_tuple(name="legend_xy", val=legend_xy, n=2, accept_none=False,
                       check_number=True, accept_none_number=True)
        ut.check_tuple(name="legend_imp_xy", val=legend_imp_xy, n=2, accept_none=False,
                       check_number=True, accept_none_number=True)
        check_imp_tuples(name="imp_ths", imp_tuples=imp_ths)
        check_imp_tuples(name="imp_marker_sizes", imp_tuples=imp_marker_sizes)
        ut.check_number_range(name="imp_bar_th", val=imp_bar_th,
                              accept_none=True, min_val=0, just_int=False)
        args_xtick = check_args_xtick(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)

        # Plot feature map
        fig, ax = plot_feature_map(df_feat=df_feat, df_cat=self._df_cat,
                                   col_cat=col_cat, col_val=col_val, col_imp=col_imp,
                                   name_test=name_test, name_ref=name_ref,
                                   figsize=figsize,
                                   start=start, **args_len, **args_seq,
                                   **args_part_color, **args_seq_color,
                                   **args_fs, weight_tmd_jmd=weight_tmd_jmd,
                                   add_xticks_pos=add_xticks_pos,
                                   grid_linewidth=grid_linewidth, grid_linecolor=grid_linecolor,
                                   border_linewidth=border_linewidth,
                                   facecolor_dark=facecolor_dark, vmin=vmin, vmax=vmax,
                                   cmap=cmap, cmap_n_colors=cmap_n_colors,
                                   cbar_pct=cbar_pct, cbar_kws=cbar_kws, cbar_xywh=cbar_xywh,
                                   dict_color=dict_color, legend_kws=legend_kws, legend_xy=legend_xy,
                                   legend_imp_xy=legend_imp_xy,
                                   imp_ths=imp_ths, imp_marker_sizes=imp_marker_sizes,
                                   imp_bar_th=imp_bar_th,
                                   **args_xtick)

        # Adjust plot
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            fig.tight_layout()
            plt.subplots_adjust(right=0.95)
        if tmd_seq is not None and seq_size is None:
            ax, seq_size = update_seq_size_(ax=ax, **args_seq, **args_part_color, **args_seq_color)
            if self._verbose:
                ut.print_out(f"Optimized sequence character fontsize is: {seq_size}")
        return fig, ax


    def update_seq_size(self,
                        ax: plt.Axes = None,
                        fig: Optional[plt.Figure] = None,
                        max_x_dist: float = 0.1,
                        fontsize_tmd_jmd: Union[int, float] = None,
                        weight_tmd_jmd: Literal['normal', 'bold'] = 'normal',
                        tmd_color: str = "mediumspringgreen",
                        jmd_color: str = "blue",
                        tmd_seq_color: str = "black",
                        jmd_seq_color: str = "white",
                        ) -> plt.Axes:
        """
        Update the font size of the sequence characters to prevent overlap.

        This method adjusts the font size of TMD-JMD sequence characters based on their provided sequences
        to ensure that the labels are clearly legible and do not overlap in the plot.

        Parameters
        ---------
        ax : plt.Axes
            CPP plot axes object.
        fig : plt.Figure, optional
            CPP plot figure object. If given, ``fontsize_tmd_jmd`` will be automatically adjusted.
        max_x_dist : float, default=0.1
            Maximum allowed horizontal distance between sequence characters during font size optimization.
            A greater value reduces potential overlaps of sequence characters.
        fontsize_tmd_jmd : int or float, optional
            Font size (>=0) for the part labels: 'JMD-N', 'TMD', 'JMD-C'. If ``None``, optimized automatically.
        weight_tmd_jmd : {'normal', 'bold'}, default='normal'
            Font weight for the part labels: 'JMD-N', 'TMD', 'JMD-C'.
        tmd_color : str, default='mediumspringgreen'
            Color for TMD.
        jmd_color : str, default='blue'
            Color for JMDs.
        tmd_seq_color : str, default='black'
            Color for TMD sequence.
        jmd_seq_color : str, default='white'
            Color for JMD sequences.

        Returns
        -------
        ax : plt.Axes
            CPP plot axes object.

        Notes
        -----
        * Use :meth:`CPPPlot.update_seq_size` AFTER :func:`matplotlib.pyplot.tight_layout`.

        See Also
        --------
        * :meth:`CPPPlot.profile` and :meth:`CPPPlot.heatmap` methods, which also use the ``tmd_seq``, ``jmd_n_seq``,
          and ``jmd_c_seq`` parameters. :meth:`CPPPlot.update_seq_size` should be called after further plot
          modifications that alter the size of figure or x-axis.

        Examples
        --------
        .. include:: examples/cpp_plot_update_seq_size.rst
        """
        # Check input
        ax = ut.check_ax(ax=ax, accept_none=False, return_first=True)
        ut.check_fig(fig=fig, accept_none=True)
        ut.check_number_range(name="max_x_dist", val=max_x_dist, min_val=0, just_int=False)
        ut.check_number_range(name="fontsize_tmd_jmd", val=fontsize_tmd_jmd, min_val=0, accept_none=True, just_int=False)
        ut.check_font_weight(name="weight_tmd_jmd", font_weight=weight_tmd_jmd)
        args_part_color = check_part_color(tmd_color=tmd_color, jmd_color=jmd_color)
        args_seq_color = check_seq_color(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
        check_match_ax_seq_len(ax=ax, jmd_c_len=self._jmd_c_len, jmd_n_len=self._jmd_n_len)
        # Adjust font size to prevent overlap
        jmd_n_seq, tmd_seq, jmd_c_seq = get_tmd_jmd_seq(ax=ax, jmd_c_len=self._jmd_c_len, jmd_n_len=self._jmd_n_len)
        args_len, args_seq = check_parts_len(jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq,
                                             jmd_c_len=self._jmd_c_len, jmd_n_len=self._jmd_n_len)
        ax, seq_size = update_seq_size_(ax=ax, **args_seq, max_x_dist=max_x_dist, **args_part_color, **args_seq_color)
        update_tmd_jmd_labels(fig=fig, seq_size=seq_size,
                              fontsize_tmd_jmd=fontsize_tmd_jmd,
                              weight_tmd_jmd=weight_tmd_jmd)
        if self._verbose:
            ut.print_out(f"Optimized sequence character fontsize is: {seq_size}")
        return ax


"""
This is a script for frontend of plotting utility function to obtain AAanalysis color list.
The backend is in general utility module to provide function to remaining AAanalysis modules.
"""
from typing import List
from aaanalysis import utils as ut


# II Main function
def plot_get_clist(n_colors: int = 3) -> List[str]:
    """
    Get a manually curated list of 2 to 9 colors or 'husl' palette for more than 9 colors.

    This functions returns one of eight different color lists optimized for appealing visualization of categories.
    If more than 9 colors are selected, :func:`seaborn.color_palette` with 'husl' palette will be used.

    Parameters
    ----------
    n_colors : int, default=3
        Number of colors. Must be greater 2.

    Returns
    -------
    list
        Color list given as matplotlib color names.

    See Also
    --------
    * The example notebooks in `Plotting Prelude <plotting_prelude.html>`_.
    * `Matplotlib color names <https://matplotlib.org/stable/gallery/color/named_colors.html>`_
    * :func:`seaborn.color_palette` function to generate a color palette in seaborn.

    Examples
    --------
    .. include:: examples/plot_get_clist.rst
    """
    # Check input
    ut.check_number_range(name="n_colors", val=n_colors, min_val=2, just_int=True)
    # Base lists
    colors = ut.plot_get_clist_(n_colors=n_colors)
    return colors





"""
This is a script for the frontend of the AAMut class for analysing the effect of amino acid changes
for given property scales."""
import aaanalysis.utils as ut


# I Helper Functions
# Check functions


# II Main Functions
class AAMut:
    """
    UNDER CONSTRUCTION - Amino Acid Mutator (AAMut) class for analyzing the impact of amino acid substitutions
    on amino acid scales.
    """
    def __init__(self, verbose=False, df_scales=None):
        self._verbose = ut.check_verbose(verbose)
        self.df_scales = df_scales

    # Main method
    def fit(self, name=None, from_aa=None, to_aa=None):
        """Compute difference between amino acids for given property scale. """

    def eval(self):
        """"""



from ._aaclust import AAclust
from ._aaclust_plot import AAclustPlot
from ._cpp_plot import CPPPlot
from ._cpp import CPP
from ._sequence_feature import SequenceFeature
from ._numerical_feature import NumericalFeature

__all__ = [
    "AAclust",
    "AAclustPlot",
    "SequenceFeature",
    "NumericalFeature",
    "CPP",
    "CPPPlot",
]